"""Change data models to accomodate for geometry-less global data model

Revision ID: 2ce65db29472
Revises: aa42863f0b44
Create Date: 2024-10-28 13:40:12.793386

"""

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "2ce65db29472"
down_revision = "aa42863f0b44"
branch_labels = None
depends_on = None


old_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    name="geodatatype",
)
new_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    "global_data",
    name="geodatatype",
)
tmp_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    "global_data",
    name="_geodatatype",
)
featuretype = sa.Enum("feature", "geo_feature", name="featuretype")

old_datatype = sa.Enum("geo_data", "generator", name="datatype")
new_datatype = sa.Enum("geo_data", "generator", "global_data", name="datatype")
tmp_datatype = sa.Enum(
    "geo_data", "generator", "global_data", name="_datatype"
)

base_data = sa.Table(
    "base_data",
    sa.MetaData(),
    sa.Column("type", new_geodatatype, nullable=True),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)

feature = sa.Table(
    "feature",
    sa.MetaData(),
    sa.Column(
        "geom",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
    sa.Column("geo_data_id", sa.Integer(), nullable=True),
    sa.Column(
        "explainability",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
)

geo_feature = sa.Table(
    "geo_feature",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "geom",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)

project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column("data_type", new_datatype, nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Replace olf geodatatype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_geo_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    old_geodatatype.drop(op.get_bind(), checkfirst=False)
    new_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_geo_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_geodatatype.drop(op.get_bind(), checkfirst=False)

    # Replace olf datatype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_datatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE _datatype"
        " USING data_type::text::_datatype"
    )
    old_datatype.drop(op.get_bind(), checkfirst=False)
    new_datatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE datatype"
        " USING data_type::text::datatype"
    )
    tmp_datatype.drop(op.get_bind(), checkfirst=False)
    op.execute(
        project_data.update()
        .where(project_data.c.data_type == None)  # noqa: E711 (using ORM)
        .values(data_type="global_data")
    )

    # Remove trigger on feature (will be replaced by trigger on geo_feature)
    op.execute(sa.DDL("DROP TRIGGER feature_make_valid ON feature;"))

    with op.batch_alter_table("base_geo_data", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_base_geo_data_extent",
            postgresql_using="gist",
            column_name="extent",
        )
    op.rename_table("base_geo_data", "base_data")
    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.alter_column(
            "type",
            type_=postgresql.ENUM(
                "geo_data",
                "generated_geo_data",
                "proposition_geo_data",
                "global_data",
                name="geodatatype",
                create_type=False,
            ),
        )
        batch_op.create_geospatial_index(
            "idx_base_data_extent",
            ["extent"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )

    op.create_geospatial_table(
        "geo_feature",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "geom",
            Geometry(
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["id"], ["feature.id"], name=op.f("fk_geo_feature_id_feature")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_geo_feature")),
    )
    with op.batch_alter_table("geo_feature", schema=None) as batch_op:
        batch_op.create_geospatial_index(
            "idx_geo_feature_geom",
            ["geom"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )

    # Populate geo_feature with all features (as they are all geo_feature)
    op.execute(
        geo_feature.insert().from_select(
            names=["id", "geom"], select=sa.select(feature.c["id", "geom"])
        )
    )

    op.create_table(
        "global_data",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("feature_id", sa.Integer(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("is_public", sa.Boolean(), nullable=False),
        sa.Column("upload_user_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["feature_id"],
            ["feature.id"],
            name=op.f("fk_global_data_feature_id_feature"),
        ),
        sa.ForeignKeyConstraint(
            ["id"], ["base_data.id"], name=op.f("fk_global_data_id_base_data")
        ),
        sa.ForeignKeyConstraint(
            ["upload_user_id"],
            ["user.id"],
            name=op.f("fk_global_data_upload_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_global_data")),
    )
    op.create_table(
        "global_data_permission",
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["global_data.id"],
            name=op.f("fk_global_data_permission_object_id_global_data"),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_global_data_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_global_data_permission")),
    )
    with op.batch_alter_table("data_attribute", schema=None) as batch_op:
        batch_op.alter_column("geo_data_id", new_column_name="data_id")
        batch_op.drop_index("ix_data_attribute_geo_data_id")
        batch_op.create_index(
            batch_op.f("ix_data_attribute_data_id"), ["data_id"], unique=False
        )
        batch_op.drop_constraint(
            "fk_data_attribute_geo_data_id_base_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_data_attribute_data_id_base_data"),
            "base_data",
            ["data_id"],
            ["id"],
        )

    # Create FeatureType enum type (else it fails)
    featuretype.create(op.get_bind(), checkfirst=False)
    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.add_column(sa.Column("type", featuretype, nullable=True))
        batch_op.alter_column("geo_data_id", new_column_name="data_id")
        batch_op.drop_geospatial_index(
            "idx_feature_geom", postgresql_using="gist", column_name="geom"
        )
        batch_op.drop_index("ix_feature_geo_data_id")
        batch_op.drop_constraint(
            "fk_feature_geo_data_id_base_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_feature_data_id_base_data"),
            "base_data",
            ["data_id"],
            ["id"],
        )
        batch_op.drop_geospatial_column("geom")
    # Only geometry feature existed before
    op.execute("UPDATE feature SET type = 'geo_feature'")

    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_geo_data_id_base_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_geo_data_id_base_data"), "base_data", ["id"], ["id"]
        )

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_project_data_data_id_base_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_project_data_data_id_base_data"),
            "base_data",
            ["data_id"],
            ["id"],
        )

    with op.batch_alter_table("zone_proposition", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_zone_proposition_data_id_base_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_zone_proposition_data_id_base_data"),
            "base_data",
            ["data_id"],
            ["id"],
        )

    # ### end Alembic commands ###
    op.execute(
        sa.DDL(
            """
/**
 * Trigger to correct the invalid geometries before an INSERT or an UPDATE.
 */
CREATE OR REPLACE FUNCTION feature_make_valid() RETURNS trigger AS $feature_make_valid$
BEGIN
    IF ST_IsValid(NEW.geom) = 'f' THEN
        NEW.geom := ST_MakeValid(NEW.geom);
    END IF;
    RETURN NEW;
END;
$feature_make_valid$ LANGUAGE plpgsql;

CREATE TRIGGER feature_make_valid BEFORE INSERT OR UPDATE ON geo_feature
    FOR EACH ROW EXECUTE PROCEDURE feature_make_valid();
    """  # noqa: E501
        )
    )
    op.execute(
        sa.DDL(
            """
/**
 * Dissolve adjacent geometries if they have the same attributes.
 *
 * Warning: the function deletes the old features before inserting the new ones.
 */
CREATE OR REPLACE FUNCTION dissolve_adjacent (data_id int) RETURNS void AS $$
BEGIN
    -- Use a temporary table to store the attributes
    CREATE TEMP TABLE temp_dissolve AS
    WITH merge AS (
        SELECT g.geom AS geom, json_object_agg(v.attribute_id, v.value) AS props
        FROM feature f INNER JOIN geo_feature g ON f.id = g.id
        INNER JOIN data_value v ON f.id = v.feature_id
        WHERE f.data_id = data_id
        GROUP BY f.id
    )
    -- Dissolve the adjacent geometries based on their attributes
    SELECT ST_SetSRID(ST_UnaryUnion(d.geom), 4326) AS geom, d.props AS props
    FROM (
        SELECT UNNEST(ST_ClusterIntersecting(m.geom)) AS geom, m.props::jsonb AS props
        FROM merge m
        GROUP BY m.props::jsonb
    ) AS d;

    -- Delete old features
    WITH delf AS (
        DELETE FROM feature f WHERE f.data_id = data_id
        RETURNING f.id
    )
    DELETE FROM geo_feature WHERE id IN (SELECT * FROM delf);

    -- Insert the new dissolved features
    WITH ins AS (
        INSERT INTO feature (data_id)
        SELECT geom, data_id
        FROM temp_dissolve
        RETURNING id, geom
    )
    INSERT INTO geo_feature (id, geom)
    SELECT id, geom FROM ins;
    INSERT INTO data_value (attribute_id, value, feature_id, type)
    SELECT d.key, d.value, ins.id, a.type
    FROM ins INNER JOIN (
        SELECT t.geom, j.key::int, j.value
        FROM temp_dissolve t, jsonb_each(t.props) j
    ) AS d ON ins.geom = d.geom
    INNER JOIN data_attribute a ON d.key = a.id;

    -- Delete the temporary table
    DROP TABLE temp_dissolve;
END;
$$ LANGUAGE plpgsql;
    """  # noqa: E501
        )
    )


def downgrade():
    op.execute(sa.DDL("DROP TRIGGER feature_make_valid ON geo_feature;"))

    # Delete all ProjectGlobalData objects
    # Need to delete all objects referencing project global data first
    op.execute(
        "DELETE FROM data_input_association pd WHERE EXISTS "
        "(SELECT 1 FROM project_data p WHERE p.id = pd.input_data_id AND "
        "p.data_type = 'global_data');"
    )
    op.execute(
        "DELETE FROM project_shared_data pd WHERE EXISTS "
        "(SELECT 1 FROM project_data p WHERE p.id = pd.project_data_id AND "
        "p.data_type = 'global_data');"
    )
    op.execute(
        project_data.delete().where(project_data.c.data_type == "global_data")
    )

    # Replace olf datatype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_datatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE _datatype"
        " USING data_type::text::_datatype"
    )
    new_datatype.drop(op.get_bind(), checkfirst=False)
    old_datatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE datatype"
        " USING data_type::text::datatype"
    )
    tmp_datatype.drop(op.get_bind(), checkfirst=False)

    # Change geodatatype enum columns
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    op.execute(
        base_data.update()
        .where(base_data.c.type == "global_data")
        .values(type=None)
    )
    tmp_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    new_geodatatype.drop(op.get_bind(), checkfirst=False)
    old_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_geodatatype.drop(op.get_bind(), checkfirst=False)
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_base_data_extent",
            postgresql_using="gist",
            column_name="extent",
        )
    op.rename_table("base_data", "base_geo_data")
    with op.batch_alter_table("base_geo_data", schema=None) as batch_op:
        batch_op.alter_column(
            "type",
            type_=postgresql.ENUM(
                "geo_data",
                "generated_geo_data",
                "proposition_geo_data",
                name="geodatatype",
                create_type=False,
            ),
        )
        batch_op.create_geospatial_index(
            "idx_base_geo_data_extent",
            ["extent"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )

    with op.batch_alter_table("zone_proposition", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_zone_proposition_data_id_base_data"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_zone_proposition_data_id_base_geo_data",
            "base_geo_data",
            ["data_id"],
            ["id"],
        )

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_project_data_data_id_base_data"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            "fk_project_data_data_id_base_geo_data",
            "base_geo_data",
            ["data_id"],
            ["id"],
        )

    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_geo_data_id_base_data"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            "fk_geo_data_id_base_geo_data", "base_geo_data", ["id"], ["id"]
        )

    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.alter_column("data_id", new_column_name="geo_data_id")
        batch_op.add_geospatial_column(
            sa.Column(
                "geom",
                Geometry(
                    spatial_index=False,
                    from_text="ST_GeomFromEWKT",
                    name="geometry",
                ),
                autoincrement=False,
                nullable=True,
            )
        )
        batch_op.drop_constraint(
            batch_op.f("fk_feature_data_id_base_data"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            "fk_feature_geo_data_id_base_geo_data",
            "base_geo_data",
            ["geo_data_id"],
            ["id"],
        )
        batch_op.create_index(
            "ix_feature_geo_data_id", ["geo_data_id"], unique=False
        )
        batch_op.create_geospatial_index(
            "idx_feature_geom",
            ["geom"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )
        batch_op.drop_column("type")

    # Populate feature geometry with all features (as they are all geo_feature)
    op.execute(
        "UPDATE feature SET geom = geo_feature.geom "
        "FROM geo_feature WHERE feature.id = geo_feature.id"
    )

    with op.batch_alter_table("data_attribute", schema=None) as batch_op:
        batch_op.alter_column("data_id", new_column_name="geo_data_id")
        batch_op.drop_constraint(
            batch_op.f("fk_data_attribute_data_id_base_data"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_data_attribute_geo_data_id_base_geo_data",
            "base_geo_data",
            ["geo_data_id"],
            ["id"],
        )
        batch_op.drop_index(batch_op.f("ix_data_attribute_data_id"))
        batch_op.create_index(
            "ix_data_attribute_geo_data_id", ["geo_data_id"], unique=False
        )

    op.drop_table("global_data_permission")
    op.drop_table("global_data")
    with op.batch_alter_table("geo_feature", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_geo_feature_geom", postgresql_using="gist", column_name="geom"
        )

    op.drop_geospatial_table("geo_feature")

    # Delete all base_geo_data without type (i.e all global data)
    # Need to delete all objects referencing global data first
    # (easier to do there, as we don't need to dereference them from base_data)
    op.execute(
        "DELETE FROM data_value WHERE feature_id IN "
        "(SELECT f.id AS feature_id FROM feature f JOIN base_geo_data g ON "
        "f.geo_data_id = g.id WHERE g.type IS null)"
    )
    op.execute(
        "DELETE FROM data_attribute WHERE geo_data_id IN "
        "(SELECT id AS geo_data_id FROM base_geo_data WHERE type IS null)"
    )
    op.execute(
        "DELETE FROM feature WHERE geo_data_id IN "
        "(SELECT id AS geo_data_id FROM base_geo_data WHERE type IS null)"
    )
    op.execute("DELETE FROM base_geo_data WHERE type IS null")

    featuretype.drop(op.get_bind(), checkfirst=False)
    # ### end Alembic commands ###

    op.execute(
        sa.DDL(
            """
/**
 * Trigger to correct the invalid geometries before an INSERT or an UPDATE.
 */
CREATE OR REPLACE FUNCTION feature_make_valid() RETURNS trigger AS $feature_make_valid$
BEGIN
    IF ST_IsValid(NEW.geom) = 'f' THEN
        NEW.geom := ST_MakeValid(NEW.geom);
    END IF;
    RETURN NEW;
END;
$feature_make_valid$ LANGUAGE plpgsql;

CREATE TRIGGER feature_make_valid BEFORE INSERT OR UPDATE ON feature
    FOR EACH ROW EXECUTE PROCEDURE feature_make_valid();
        """  # noqa: E501
        )
    )
    op.execute(
        sa.DDL(
            """
/**
 * Dissolve adjacent geometries if they have the same attributes.
 *
 * Warning: the function deletes the old features before inserting the new ones.
 */
CREATE OR REPLACE FUNCTION dissolve_adjacent (data_id int) RETURNS void AS $$
BEGIN
    -- Use a temporary table to store the attributes
    CREATE TEMP TABLE temp_dissolve AS
    WITH merge AS (
        SELECT f.geom AS geom, json_object_agg(v.attribute_id, v.value) AS props
        FROM feature f INNER JOIN data_value v ON f.id = v.feature_id
        WHERE f.geo_data_id = data_id
        GROUP BY f.id
    )
    -- Dissolve the adjacent geometries based on their attributes
    SELECT ST_SetSRID(ST_UnaryUnion(d.geom), 4326) AS geom, d.props AS props
    FROM (
        SELECT UNNEST(ST_ClusterIntersecting(m.geom)) AS geom, m.props::jsonb AS props
        FROM merge m
        GROUP BY m.props::jsonb
    ) AS d;

    -- Delete old features
    DELETE FROM feature WHERE geo_data_id = data_id;

    -- Insert the new dissolved features
    WITH ins AS (
        INSERT INTO feature (geom, geo_data_id)
        SELECT geom, data_id
        FROM temp_dissolve
        RETURNING id, geom
    )
    INSERT INTO data_value (attribute_id, value, feature_id, type)
    SELECT d.key, d.value, ins.id, a.type
    FROM ins INNER JOIN (
        SELECT t.geom, j.key::int, j.value
        FROM temp_dissolve t, jsonb_each(t.props) j
    ) AS d ON ins.geom = d.geom
    INNER JOIN data_attribute a ON d.key = a.id;

    -- Delete the temporary table
    DROP TABLE temp_dissolve;
END;
$$ LANGUAGE plpgsql;
        """  # noqa: E501
        )
    )



from marshmallow import fields

from ..models import DataShare
from .base import BaseSchema
from .geo_data import BaseDataSchema


class DataShareSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        # This schema is used for creation/retrieval only
        model = DataShare
        fields = (
            "id",
            "uid",
            "expiration",
            "expired",
            "data",
        )
        dump_only = ("id", "uid", "data")

    data = fields.Nested(BaseDataSchema, only=("id", "type"))



DEBUG = False

PROPAGATE_EXCEPTIONS = True  # allow the app to handle exceptions

# SQLAlchemy configuration
SQLALCHEMY_DATABASE_URI = ""  # overwrite this option in `instance/config.py`
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Swagger configuration (for flasgger)
SWAGGER = {"title": "DESEASION API docs", "openapi": "3.0.2"}



from datetime import datetime

from celery import Task, shared_task, states
from celery.result import AsyncResult
from celery.signals import (
    task_failure,
    task_postrun,
    task_prerun,
    task_revoked,
)

from ..exceptions import ProcessingError
from ..models import (
    DataStream,
    GeoDataType,
    ProcessingModel,
    ProjectTaskModel,
    ProjectTaskType,
    db,
)
from ..models.geo_data_utils import ModelProcessingService
from ..services.geo_data_loading_service import load_from_wfs, load_from_wms


class ContextTask(Task):
    app_context = None

    def __call__(self, *args, **kwargs):
        with self.app_context:
            return self.run(*args, **kwargs)


class ProgressFeedback:
    def __init__(self, task):
        self._task = task
        self.message = None

    def set_progress(self, count, total):
        progress = int(100 * count / total)
        self._task.update_state(
            state=states.STARTED,
            meta={"progress": progress, "message": self.message},
        )

    def set_message(self, message):
        self.message = message
        self._task.update_state(
            state=states.STARTED, meta={"message": message}
        )

    def forget_message(self):
        self.message = None
        self._task.update_state(state=states.STARTED, meta={"message": None})


@shared_task(name=ProjectTaskType.process_project_data.name, bind=True)
def process_and_set_model(self, *, model_id):
    """Process the model and save the new data."""
    model = ProcessingModel.get_by_id(model_id)
    processing_service = ModelProcessingService()
    feedback = ProgressFeedback(self)
    with db.session.no_autoflush:
        features = processing_service.process_model(model, feedback=feedback)
        processing_service.set_project_data_features(
            model.data_generator, features
        )


@shared_task(name=ProjectTaskType.update_stream.name, bind=True)
def update_stream_data(
    *args,
    data_id,
):
    """Update the DataStream data field."""
    data_stream = DataStream.get_by_id(data_id)
    processing_service = ModelProcessingService()
    stream = data_stream.stream
    with db.session.no_autoflush:
        if stream.type == GeoDataType.wfs:
            attributes = (
                data_stream.data.attributes if data_stream.data else None
            )
            features = load_from_wfs(stream, attributes)
        else:
            attribute = None
            if data_stream.data and len(data_stream.data.attributes) == 1:
                attribute = data_stream.data.attributes[0]
            features = load_from_wms(
                stream,
                classes=data_stream.classes,
                start=data_stream.start,
                step=data_stream.step,
                stop=data_stream.stop,
                resolution=data_stream.resolution,
                old_attribute=attribute,
            )
        # Keep features with only None values (it's an input, we keep it as is)
        processing_service.set_project_data_features(
            data_stream, features, filter_none=False
        )


@task_prerun.connect
def project_task_prerun(task_id, task, args, kwargs, **kw):
    """Prerun signal for the tasks."""
    try:
        task_type = ProjectTaskType[task.name]
    except KeyError:
        return
    time = datetime.utcnow()
    with task.app_context:
        project = None
        if task_type == ProjectTaskType.process_project_data:
            model_id = kwargs["model_id"]
            model = ProcessingModel.get_by_id(model_id)
            project = model.data_generator.project
        elif task_type == ProjectTaskType.update_stream:
            data_stream_id = kwargs["data_id"]
            data_stream = DataStream.get_by_id(data_stream_id)
            project = data_stream.project
        if project is not None:
            task_model = ProjectTaskModel(
                task_id=task_id, project=project, started_at=time
            )
            task_model.type = task_type
            task_model.params = kwargs
            task_model.state = states.STARTED
            task_model.create()
    print("task {} ({}) started at {}".format(task.name, task_id, time))


@task_postrun.connect
def project_task_postrun(task_id, task, args, kwargs, retval, state, **kw):
    """Postrun signal for the tasks

    If the task name is 'process_project_data', the final status and finish
    time is saved to the database.
    """
    if state is None:
        result = AsyncResult(task_id)
        state = result.state
    time = datetime.utcnow()
    with task.app_context:  # use the flask application context
        task_model = ProjectTaskModel.get_by_task_id(task_id)
        if task_model is not None:
            task_model.finished_at = time
            task_model.state = state
            task_model.update()
    print(
        "task {} ({}) finishing with state {} at {}".format(
            task.name, task_id, state, time
        )
    )


@task_failure.connect
def project_task_failure(
    task_id, exception, args, kwargs, traceback, einfo, **kw
):
    """Save the 'FAILED' state and the error message when a task fails"""
    if isinstance(exception, ProcessingError):
        with process_and_set_model.app_context:
            task_model = ProjectTaskModel.get_by_task_id(task_id)
            if task_model is not None:
                task_model.error_message = str(exception)
                task_model.finished_at = datetime.utcnow()
                task_model.update()


@task_revoked.connect
def project_task_revoked(request, terminated, signum, expired, **wk):
    """Save the 'REVOKED' state and the time when a task is revoked"""
    with request.task.app_context:
        task_model = ProjectTaskModel.get_by_task_id(request.task_id)
        if task_model is not None:
            task_model.finished_at = datetime.utcnow()
            task_model.state = states.REVOKED
            task_model.update()



import subprocess as sp

import simplejson as json
from flask import current_app as app

from ..exceptions import ProcessingError


def execute_sandbox(script_name, json_input=None):
    """Execute the script in a sandbox.

    Args:
        script_name -- The filename of the script.
        json_input -- The data the use as input in the script.
    """
    # TODO: Whitelist the modules the user is allowed to import in the rules
    # TODO: Add a time out for the execution of the user's code
    # TODO: Limit the memory allowed for the user's code execution
    # TODO: Prevent the user to writing to stdout
    p = sp.Popen(
        [
            "docker",
            "run",
            "--rm",
            "-i",
            "registry.gitlab.com/decide.imt-atlantique/deseasion/sandbox",
            script_name,
        ],
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
    )
    result, _ = p.communicate(input=json_input.encode())
    return result


def load_json_bytes(bytes_data):
    """Loads a json bytes object.

    Raises:
        ProcessingError if the json data has an 'error' key or an unknown key.
    """
    if not bytes_data:
        raise ProcessingError("Unknown error in the data rule")
    try:
        result = json.loads(bytes_data.decode())
    except ValueError as err:
        app.logger.info("Error while parsing text: {}".format(bytes_data))
        raise err
    if "error" in result:
        raise ProcessingError(result["error"])
    elif "data" in result:
        return result["data"]
    else:
        raise ProcessingError(
            "Unknown error during the processing of the data"
        )


def evaluate_discrete_rule(data, rule, stats=None):
    """Evaluate the discrete model rule with the data.

    Args:
        data (list): A JSON-serializable list of dict to use as the input for
            the rules.
        rule (str): A Python code.

    Returns:
        A list of booleans.
    """
    data_dict = {"data": data, "code": rule}
    if stats is not None:
        data_dict["statistics"] = stats
    json_input = json.dumps(data_dict, ignore_nan=True)
    result = execute_sandbox("process_category.py", json_input)
    result_list = load_json_bytes(result)
    # Verify that all the processed values are boolean
    if not all([isinstance(value, bool) for value in result_list]):
        raise ProcessingError(
            "The result of a rule evaluation should be a boolean"
        )
    return [value is True for value in result_list]


def evaluate_continuous_rule(data, rule, stats=None):
    """Evaluate the continuous model rule with the data.

    Args:
        data (list): A JSON-serializable list of dict to use as the input for
            the rules.
        rule (str): A Python code.
        stats (dict): A dictionary with the statistics on the global attributes

    Returns:
        A list of dict.
    """
    data_dict = {"data": data, "code": rule}
    if stats is not None:
        data_dict["statistics"] = stats
    json_input = json.dumps(data_dict, ignore_nan=True)
    result = execute_sandbox("process_rule.py", json_input)
    result_list = load_json_bytes(result)
    return result_list



"""Make streams base data

Revision ID: bb99ff70c7bf
Revises: 163a9f52cac3
Create Date: 2025-06-25 17:03:17.087567

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "bb99ff70c7bf"
down_revision = "163a9f52cac3"
branch_labels = None
depends_on = None


old_data_type = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "global_data",
    name="geodatatype",
)
new_data_type = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "global_data",
    "wfs",
    "wms",
    name="geodatatype",
)
tmp_data_type = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "global_data",
    "wfs",
    "wms",
    name="_geodatatype",
)
stream_type = sa.Enum(
    "wfs",
    "wms",
    name="streamdatatype",
)


def upgrade():
    # Replace old data type with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    old_data_type.drop(op.get_bind(), checkfirst=False)
    new_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_data_type.drop(op.get_bind(), checkfirst=False)

    op.create_table(
        "base_data_permission",
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["base_data.id"],
            name=op.f("fk_base_data_permission_object_id_base_data"),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_base_data_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_base_data_permission")),
    )
    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("original_name", sa.String(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("description", sa.String(), nullable=True)
        )
        # Make it non-nullable afterwards (necessary to create column)
        batch_op.add_column(
            sa.Column("is_public", sa.Boolean(), nullable=True)
        )

    # Move stored streams in temporary tables
    op.execute(
        sa.text(
            """
CREATE TEMPORARY TABLE temp_wfs_objects AS
SELECT s.id, s.name, s.title, s.original_name, s.url,
s.upload_user_id, s.description, s.keywords, s.created_at,
s.modified_at, s.is_public, w.feature_type, w.version,
0 as new_id
FROM stream_geo_data s JOIN wfs_geo_data w ON s.id = w.id
"""
        )
    )
    op.execute(
        sa.text(
            "CREATE TEMPORARY TABLE temp_wms_objects AS "
            "SELECT s.id, s.name, s.title, s.original_name, s.url, "
            "s.upload_user_id, s.description, s.keywords, s.created_at, "
            "s.modified_at, s.is_public, w.layer, w.version, w.classes, "
            "w.start, w.step, w.stop, w.resolution, 0 as new_id "
            "FROM stream_geo_data s JOIN wms_geo_data w ON s.id = w.id"
        )
    )

    # Clear stream tables
    op.execute(sa.text("DELETE FROM wfs_geo_data"))
    op.execute(sa.text("DELETE FROM wms_geo_data"))

    with op.batch_alter_table("wfs_geo_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("upload_user_id", sa.Integer(), nullable=True)
        )
        batch_op.add_column(sa.Column("title", sa.String(), nullable=True))
        batch_op.add_column(sa.Column("url", sa.String(), nullable=False))
        batch_op.add_column(
            sa.Column(
                "keywords",
                postgresql.ARRAY(sa.String(), dimensions=1),
                nullable=True,
            )
        )
        batch_op.drop_constraint(
            "fk_wfs_geo_data_id_stream_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_wfs_geo_data_upload_user_id_user"),
            "user",
            ["upload_user_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_wfs_geo_data_id_base_data"),
            "base_data",
            ["id"],
            ["id"],
        )

    with op.batch_alter_table("wms_geo_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("upload_user_id", sa.Integer(), nullable=True)
        )
        batch_op.add_column(sa.Column("title", sa.String(), nullable=True))
        batch_op.add_column(sa.Column("url", sa.String(), nullable=False))
        batch_op.add_column(
            sa.Column(
                "keywords",
                postgresql.ARRAY(sa.String(), dimensions=1),
                nullable=True,
            )
        )
        batch_op.drop_constraint(
            "fk_wms_geo_data_id_stream_geo_data", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_wms_geo_data_id_base_data"),
            "base_data",
            ["id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_wms_geo_data_upload_user_id_user"),
            "user",
            ["upload_user_id"],
            ["id"],
        )

    # Insert WFS streams into base_data and store id in temp_wfs_objects.new_id
    op.execute(
        sa.text(
            """
WITH numbered_source AS (
    SELECT
        id AS temp_id,
        name,
        original_name,
        description,
        created_at,
        modified_at,
        is_public,
        ROW_NUMBER() OVER () AS rn
    FROM temp_wfs_objects
),
inserted AS (
    INSERT INTO base_data (
        type, name, original_name, description, created_at, modified_at, is_public
    )
    SELECT
        'wfs', name, original_name, description, created_at, modified_at, is_public
    FROM numbered_source
    RETURNING id
),
numbered_inserted AS (
    SELECT
        id AS new_id,
        ROW_NUMBER() OVER () AS rn
    FROM inserted
)
UPDATE temp_wfs_objects t
SET new_id = ni.new_id
FROM numbered_source s
JOIN numbered_inserted ni ON s.rn = ni.rn
WHERE t.id = s.temp_id;
"""  # noqa: E501
        )
    )
    # Insert WFS stream parts in wfs_geo_data
    op.execute(
        sa.text(
            """
INSERT INTO wfs_geo_data (id, title, url, keywords, version, feature_type)
SELECT new_id, title, url, keywords, version, feature_type FROM temp_wfs_objects
"""  # noqa: E501
        )
    )

    # Insert WMS streams into base_data and store id in temp_wms_objects.new_id
    op.execute(
        sa.text(
            """
WITH numbered_source AS (
    SELECT
        id AS temp_id,
        name,
        original_name,
        description,
        created_at,
        modified_at,
        is_public,
        ROW_NUMBER() OVER () AS rn
    FROM temp_wms_objects
),
inserted AS (
    INSERT INTO base_data (
        type, name, original_name, description, created_at, modified_at, is_public
    )
    SELECT
        'wfs', name, original_name, description, created_at, modified_at, is_public
    FROM numbered_source
    RETURNING id
),
numbered_inserted AS (
    SELECT
        id AS new_id,
        ROW_NUMBER() OVER () AS rn
    FROM inserted
)
UPDATE temp_wms_objects t
SET new_id = ni.new_id
FROM numbered_source s
JOIN numbered_inserted ni ON s.rn = ni.rn
WHERE t.id = s.temp_id;
"""  # noqa: E501
        )
    )
    # Insert WMS stream parts in wms_geo_data
    op.execute(
        sa.text(
            """
INSERT INTO wms_geo_data (id, title, url, keywords, version, classes,
start, step, stop, layer, resolution)
SELECT new_id, title, url, keywords, version, classes,
start, step, stop, layer, resolution
FROM temp_wms_objects
"""
        )
    )

    # Move migrated geo_data fields values to base_data
    op.execute(
        sa.text(
            """
UPDATE base_data
SET
    description = other_table.description,
    original_name = other_table.original_name,
    is_public = other_table.is_public
FROM (
    SELECT id, description, original_name, is_public
    FROM geo_data
) AS other_table
WHERE other_table.id = base_data.id
"""
        )
    )
    # Move migrated global_data fields values to base_data
    op.execute(
        sa.text(
            """
UPDATE base_data
SET
    description = other_table.description,
    is_public = other_table.is_public
FROM (
    SELECT id, description, is_public
    FROM global_data
) AS other_table
WHERE other_table.id = base_data.id
"""
        )
    )
    # Set is_public to false for generated_geo_data
    op.execute(
        sa.text(
            """
UPDATE base_data
SET
    is_public = false
WHERE base_data.type = 'generated_geo_data'
"""
        )
    )
    op.alter_column("base_data", "is_public", nullable=False)
    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.drop_column("description")
        batch_op.drop_column("is_public")
        batch_op.drop_column("original_name")

    with op.batch_alter_table("global_data", schema=None) as batch_op:
        batch_op.drop_column("description")
        batch_op.drop_column("is_public")

    # Add WFS stream permissions
    op.execute(
        sa.text(
            """
INSERT INTO base_data_permission (object_id, user_id)
SELECT w.new_id, p.user_id FROM stream_geo_data_permission p
JOIN temp_wfs_objects w ON p.object_id = w.id
"""
        )
    )
    # Add WMS stream permissions
    op.execute(
        sa.text(
            """
INSERT INTO base_data_permission (object_id, user_id)
SELECT w.new_id, p.user_id FROM stream_geo_data_permission p
JOIN temp_wms_objects w ON p.object_id = w.id
"""
        )
    )
    # Add geo data permissions
    op.execute(
        sa.text(
            """
INSERT INTO base_data_permission (object_id, user_id)
SELECT object_id, user_id FROM geo_data_permission
"""
        )
    )
    # Add global data permissions
    op.execute(
        sa.text(
            """
INSERT INTO base_data_permission (object_id, user_id)
SELECT object_id, user_id FROM global_data_permission
"""
        )
    )
    op.drop_table("stream_geo_data_permission")
    op.drop_table("geo_data_permission")
    op.drop_table("global_data_permission")

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_project_data_stream_id_stream_geo_data", type_="foreignkey"
        )
    # Modify data stream in project data to reference base data (stream ids)
    op.execute(
        sa.text(
            """
UPDATE project_data
SET
    stream_id = other.new_id
FROM (
    SELECT 'wfs' AS type, id, new_id FROM temp_wfs_objects UNION
    SELECT 'wms' AS type, id, new_id FROM temp_wms_objects
) AS other
WHERE project_data.stream_id = other.id
"""
        )
    )
    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_project_data_stream_id_base_data"),
            "base_data",
            ["stream_id"],
            ["id"],
        )

    op.drop_table("stream_geo_data")
    stream_type.drop(op.get_bind(), checkfirst=False)


def downgrade():
    op.create_table(
        "stream_geo_data",
        sa.Column(
            "type",
            postgresql.ENUM("wfs", "wms", name="streamdatatype"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("name", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column("title", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "original_name", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column("url", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "upload_user_id", sa.INTEGER(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "keywords",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "is_public", sa.BOOLEAN(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "modified_at",
            postgresql.TIMESTAMP(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "id",
            sa.INTEGER(),
            server_default=sa.text(
                "nextval('stream_geo_data_id_seq'::regclass)"
            ),
            autoincrement=True,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["upload_user_id"],
            ["user.id"],
            name="fk_stream_geo_data_upload_user_id_user",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_stream_geo_data"),
        postgresql_ignore_search_path=False,
    )

    # Move stored streams in temporary tables
    op.execute(
        sa.text(
            "CREATE TEMPORARY TABLE temp_wfs_objects AS "
            "SELECT d.id, d.name, w.title, d.original_name, w.url, "
            "w.upload_user_id, d.description, w.keywords, d.created_at, "
            "d.modified_at, d.is_public, w.feature_type, w.version, "
            "0 as new_id "
            "FROM base_data d JOIN wfs_geo_data w ON d.id = w.id "
        )
    )
    op.execute(
        sa.text(
            "CREATE TEMPORARY TABLE temp_wms_objects AS "
            "SELECT d.id, d.name, w.title, d.original_name, w.url, "
            "w.upload_user_id, d.description, w.keywords, d.created_at, "
            "d.modified_at, d.is_public, w.layer, w.version, w.classes, "
            "w.start, w.step, w.stop, w.resolution, 0 as new_id "
            "FROM base_data d JOIN wms_geo_data w ON d.id = w.id"
        )
    )

    # Clear stream tables
    op.execute(sa.text("DELETE FROM wfs_geo_data"))
    op.execute(sa.text("DELETE FROM wms_geo_data"))

    with op.batch_alter_table("wms_geo_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_wms_geo_data_upload_user_id_user"),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("fk_wms_geo_data_id_base_data"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            "fk_wms_geo_data_id_stream_geo_data",
            "stream_geo_data",
            ["id"],
            ["id"],
        )
        batch_op.drop_column("keywords")
        batch_op.drop_column("url")
        batch_op.drop_column("title")
        batch_op.drop_column("upload_user_id")

    with op.batch_alter_table("wfs_geo_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_wfs_geo_data_id_base_data"), type_="foreignkey"
        )
        batch_op.drop_constraint(
            batch_op.f("fk_wfs_geo_data_upload_user_id_user"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_wfs_geo_data_id_stream_geo_data",
            "stream_geo_data",
            ["id"],
            ["id"],
        )
        batch_op.drop_column("keywords")
        batch_op.drop_column("url")
        batch_op.drop_column("title")
        batch_op.drop_column("upload_user_id")

    # Insert WFS streams into base_data and store id in temp_wfs_objects.new_id
    op.execute(
        sa.text(
            """
WITH numbered_source AS (
    SELECT
        id AS temp_id,
        name,
        title,
        original_name,
        description,
        created_at,
        modified_at,
        is_public,
        url,
        upload_user_id,
        keywords,
        ROW_NUMBER() OVER () AS rn
    FROM temp_wfs_objects
),
inserted AS (
    INSERT INTO stream_geo_data (
        type, name, title, original_name, description, created_at, modified_at,
        is_public, url, upload_user_id, keywords
    )
    SELECT
        'wfs', name, title, original_name, description, created_at, modified_at,
        is_public, url, upload_user_id, keywords
    FROM numbered_source
    RETURNING id
),
numbered_inserted AS (
    SELECT
        id AS new_id,
        ROW_NUMBER() OVER () AS rn
    FROM inserted
)
UPDATE temp_wfs_objects t
SET new_id = ni.new_id
FROM numbered_source s
JOIN numbered_inserted ni ON s.rn = ni.rn
WHERE t.id = s.temp_id;
"""  # noqa: E501
        )
    )
    # Insert WFS stream parts in wfs_geo_data
    op.execute(
        sa.text(
            """
INSERT INTO wfs_geo_data (id, feature_type, version)
SELECT new_id, feature_type, version FROM temp_wfs_objects
"""  # noqa: E501
        )
    )

    # Insert WMS streams into base_data and store id in temp_wfs_objects.new_id
    op.execute(
        sa.text(
            """
WITH numbered_source AS (
    SELECT
        id AS temp_id,
        name,
        title,
        original_name,
        description,
        created_at,
        modified_at,
        is_public,
        url,
        upload_user_id,
        keywords,
        ROW_NUMBER() OVER () AS rn
    FROM temp_wms_objects
),
inserted AS (
    INSERT INTO stream_geo_data (
        type, name, title, original_name, description, created_at, modified_at,
        is_public, url, upload_user_id, keywords
    )
    SELECT
        'wfs', name, title, original_name, description, created_at, modified_at,
        is_public, url, upload_user_id, keywords
    FROM numbered_source
    RETURNING id
),
numbered_inserted AS (
    SELECT
        id AS new_id,
        ROW_NUMBER() OVER () AS rn
    FROM inserted
)
UPDATE temp_wms_objects t
SET new_id = ni.new_id
FROM numbered_source s
JOIN numbered_inserted ni ON s.rn = ni.rn
WHERE t.id = s.temp_id;
"""  # noqa: E501
        )
    )
    # Insert WMS stream parts in wms_geo_data
    op.execute(
        sa.text(
            """
INSERT INTO wms_geo_data (id, version, classes,
start, step, stop, layer, resolution)
SELECT new_id, version, classes,
start, step, stop, layer, resolution
FROM temp_wms_objects
"""
        )
    )

    with op.batch_alter_table("global_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "is_public", sa.BOOLEAN(), autoincrement=False, nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "description", sa.VARCHAR(), autoincrement=False, nullable=True
            )
        )

    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "original_name",
                sa.VARCHAR(),
                autoincrement=False,
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column(
                "is_public", sa.BOOLEAN(), autoincrement=False, nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "description", sa.VARCHAR(), autoincrement=False, nullable=True
            )
        )

    # Move migrated geo_data fields values back from base_data
    op.execute(
        sa.text(
            """
UPDATE geo_data
SET
    description = other_table.description,
    original_name = other_table.original_name,
    is_public = other_table.is_public
FROM (
    SELECT id, description, original_name, is_public
    FROM base_data
) AS other_table
WHERE other_table.id = geo_data.id
"""
        )
    )
    # Move migrated global_data fields values back from base_data
    op.execute(
        sa.text(
            """
UPDATE global_data
SET
    description = other_table.description,
    is_public = other_table.is_public
FROM (
    SELECT id, description, is_public
    FROM base_data
) AS other_table
WHERE other_table.id = global_data.id
"""
        )
    )

    op.alter_column("geo_data", "is_public", nullable=False)
    op.alter_column("global_data", "is_public", nullable=False)
    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.drop_column("is_public")
        batch_op.drop_column("description")
        batch_op.drop_column("original_name")

    op.create_table(
        "geo_data_permission",
        sa.Column(
            "object_id", sa.INTEGER(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "user_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["geo_data.id"],
            name="fk_geo_data_permission_object_id_geo_data",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_geo_data_permission_user_id_user",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_geo_data_permission"),
    )
    op.create_table(
        "global_data_permission",
        sa.Column(
            "object_id", sa.INTEGER(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "user_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["global_data.id"],
            name="fk_global_data_permission_object_id_global_data",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_global_data_permission_user_id_user",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_global_data_permission"),
    )
    op.create_table(
        "stream_geo_data_permission",
        sa.Column(
            "object_id", sa.INTEGER(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "user_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["stream_geo_data.id"],
            name="fk_stream_geo_data_permission_object_id_stream_geo_data",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_stream_geo_data_permission_user_id_user",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_stream_geo_data_permission"),
    )

    # Add WFS stream permissions
    op.execute(
        sa.text(
            """
INSERT INTO stream_geo_data_permission (object_id, user_id)
SELECT w.new_id, p.user_id FROM base_data_permission p
JOIN temp_wfs_objects w ON p.object_id = w.id
"""
        )
    )
    # Add WMS stream permissions
    op.execute(
        sa.text(
            """
INSERT INTO stream_geo_data_permission (object_id, user_id)
SELECT w.new_id, p.user_id FROM base_data_permission p
JOIN temp_wms_objects w ON p.object_id = w.id
"""
        )
    )
    # Add geo data permissions
    op.execute(
        sa.text(
            """
INSERT INTO geo_data_permission (object_id, user_id)
SELECT object_id, user_id FROM base_data_permission
JOIN base_data ON base_data.id = base_data_permission.object_id
WHERE base_data.type = 'geo_data'
"""
        )
    )
    # Add global data permissions
    op.execute(
        sa.text(
            """
INSERT INTO global_data_permission (object_id, user_id)
SELECT object_id, user_id FROM base_data_permission
JOIN base_data ON base_data.id = base_data_permission.object_id
WHERE base_data.type = 'global_data'
"""
        )
    )

    op.drop_table("base_data_permission")

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_project_data_stream_id_base_data"),
            type_="foreignkey",
        )

    # Modify data stream in project data to reference stream geo data
    op.execute(
        sa.text(
            """
UPDATE project_data
SET
    stream_id = other.new_id
FROM (
    SELECT 'wfs' AS type, id, new_id FROM temp_wfs_objects UNION
    SELECT 'wms' AS type, id, new_id FROM temp_wms_objects
) AS other
WHERE project_data.stream_id = other.id
"""
        )
    )

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_project_data_stream_id_stream_geo_data",
            "stream_geo_data",
            ["stream_id"],
            ["id"],
        )

    # Delete stream from base_data table
    op.execute(sa.text("DELETE FROM base_data WHERE type IN ('wfs', 'wms')"))
    # Replace new data type with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    new_data_type.drop(op.get_bind(), checkfirst=False)
    old_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_data_type.drop(op.get_bind(), checkfirst=False)



import functools
from typing import Type

from flask import jsonify, request
from marshmallow import Schema

from ..exceptions import PermissionError, RequestError
from ..services.permission_service import has_permission, has_permission_for_id


def get_json_content(exception=True):
    content = request.get_json()
    if content is None and exception:
        raise RequestError(
            "Invalid content: Accepting application/json type", 400
        )
    return content


def dump_data(serializer, **kwargs):
    dict_data = {}
    for k, v in kwargs.items():
        serialized = serializer.dump(v)
        dict_data[k] = serialized
    return jsonify(**dict_data)


def check_permission(obj):
    """Check if the user of the request has permissions on the object

    Raises:
        PermissionError if the user does not have permissions
    """
    if not has_permission(obj):
        raise PermissionError


def check_permission_for_id(obj_class, obj_id):
    """Check if the user of the request has permissions on the object with the
    given id

    Raises:
        PermissionError if the user does not have permissions
    """
    if not has_permission_for_id(obj_class, obj_id):
        raise PermissionError


_openapi_types = {"int": "integer", "bool": "boolean", "str": "string"}


def type2openapi(type_: type | str) -> str:
    """Convert python type to openAPI compliant type.

    :param type_:
    :return: openAPI type
    """
    _type = type_.__name__ if isinstance(type_, type) else type_
    return _openapi_types.get(_type, _type)


def with_request_body(
    schema: Schema | Type[Schema] | list[Schema | Type[Schema]] | str = None,
    description: str = "",
    required: bool = True,
    content: dict = None,
):
    """Annotate operation function with openAPI requestBody.

    If `schema` is set, it will add an application/json content with it as
    schema.

    :param schema:
        either a schema or schema instance (for one schema) or a list of
        schema instances or classes (for polymorphism with response type).
        Can also be a string that will be used 'as is' for the schema
        reference.
    :param description:
    :param required:
    :param content:
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        wrapped.__apispec__ = getattr(func, "__apispec__", {})
        wrapped.__apispec__["requestBody"] = {
            "schema": schema,
            "description": description,
            "required": required,
            "content": content or {},
        }
        return wrapped

    return decorate


def with_response(
    status: int,
    schema: Schema | Type[Schema] | list[Schema | Type[Schema]] | str = None,
    description: str = "",
    content: dict | None = None,
):
    """Annotate operation function with openAPI response.

    If `schema` is set, it will add an application/json content with it as
    schema.

    :param status:
    :param schema:
        either a schema or schema instance (for one schema) or a list of
        schema instances or classes (for polymorphism with response type).
        Can also be a string that will be used 'as is' for the schema
        reference.
    :param description:
    :param content:
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        wrapped.__apispec__ = getattr(func, "__apispec__", {})
        wrapped.__apispec__["responses"] = wrapped.__apispec__.get(
            "responses", {}
        )
        wrapped.__apispec__["responses"][status] = {
            "schema": schema,
            "description": description,
            "content": content or {},
        }
        return wrapped

    return decorate


def with_query_arg(
    name: str,
    type_: str | type | None = None,
    required: bool = True,
    description: str = "",
    **kwargs,
):
    """Annotate operation function with openAPI parameters for query args.

    :param name:
    :param type_:
    :param required:
    :param description:

    .. note::
        any additional keyword argument is added as is in the openAPI parameter
        created by this decorator
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        wrapped.__apispec__ = getattr(func, "__apispec__", {})
        wrapped.__apispec__["parameters"] = wrapped.__apispec__.get(
            "parameters", []
        )
        parameter = {
            "in": "query",
            "name": name,
            "required": required,
            "description": description,
        }
        if type_ is not None:
            parameter["schema"] = {"type": type2openapi(type_)}
        parameter = {**parameter, **kwargs}
        wrapped.__apispec__["parameters"].append(parameter)
        return wrapped

    return decorate



from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import backref
from sqlalchemy.orm.collections import attribute_mapped_collection

from . import db
from .mixins import BaseModelMixin, ModelMixin


class MRSortInferenceAlternative(ModelMixin, db.Model):
    """Alternatives for the inference of the MR-Sort model.

    Attributes:
        category: The assigned category.
        visible: Visible in the list of inference alternatives.
        mrsort: MR-Sort model.
    """

    __tablename__ = "mrsort_inference_alternative"

    category_id = db.Column(
        db.Integer, db.ForeignKey("discrete_category.id", ondelete="SET NULL")
    )
    mrsort_id = db.Column(
        db.Integer,
        db.ForeignKey("processing_model.id", ondelete="CASCADE"),
        nullable=False,
    )

    mrsort = db.relationship(
        "MRSort",
        backref=backref("inference_alternatives", cascade="all,delete-orphan"),
    )
    category = db.relationship("DiscreteCategory")
    inference_values = db.relationship(
        "MRSortInferenceValue",
        collection_class=attribute_mapped_collection("criterion"),
        cascade="all,delete-orphan",
        back_populates="alternative",
    )
    values = association_proxy(
        "inference_values",
        "value",
        creator=lambda k, v: MRSortInferenceValue(criterion=k, value=v),
    )


class MRSortInferenceValue(BaseModelMixin, db.Model):
    __tablename__ = "mrsort_inference_value"

    mrsort_id = db.Column(db.Integer, primary_key=True)
    attribute_id = db.Column(db.Integer, primary_key=True)
    alternative_id = db.Column(
        db.Integer,
        db.ForeignKey("mrsort_inference_alternative.id", ondelete="CASCADE"),
        primary_key=True,
    )
    value = db.Column(JSONB)

    __table_args__ = (
        db.ForeignKeyConstraint(
            ["mrsort_id", "attribute_id"],
            ["mrsort_criterion.mrsort_id", "mrsort_criterion.attribute_id"],
            ondelete="CASCADE",
        ),
    )

    criterion = db.relationship(
        "MRSortCriterion",
        backref=backref("inference_values", cascade="delete,delete-orphan"),
    )
    alternative = db.relationship(
        "MRSortInferenceAlternative", back_populates="inference_values"
    )



import math
import random
from functools import partial

import numpy as np
import pyproj
from shapely.geometry import Polygon
from shapely.ops import transform
from shapely.strtree import STRtree


class GAContext:
    def __init__(self, df, features, geo_size, bounds=None, proj=""):
        self.df = df
        self.features = features
        self.geo_size = geo_size
        self.proj = proj
        if bounds is None:
            self.bounds = tuple(df.total_bounds)
        else:
            self.bounds = bounds

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, value):
        self._df = value
        geoms = []
        values = []
        for i, f in self._df.iterrows():
            values.append(f["value"])
            geoms.append(f.geometry)
        self._geometries = geoms
        self._geometries_tree = STRtree(geoms)
        self._values = values

    @property
    def values(self):
        return self._values

    @property
    def geometries(self):
        return self._geometries

    @property
    def geometries_tree(self):
        return self._geometries_tree


class Individual:
    def __init__(self, ctx, genes):
        self._ctx = ctx
        self.genes = genes

    @classmethod
    def random(cls, ctx):
        """Create a new random individual."""
        xmin, ymin, xmax, ymax = ctx.bounds
        dmin = 0.5 * math.sqrt(ctx.geo_size)
        dmax = 1.5 * math.sqrt(ctx.geo_size)
        alpha_min = 0.1

        x = random.uniform(xmin, xmax)
        y = random.uniform(ymin, ymax)
        alpha1 = random.uniform(alpha_min, math.pi / 2 - alpha_min)
        d1 = random.uniform(dmin, dmax)
        alpha2 = random.uniform(alpha_min, math.pi / 2 - alpha_min)
        d2 = random.uniform(dmin, dmax)
        alpha3 = random.uniform(alpha_min, math.pi / 2 - alpha_min)
        d3 = random.uniform(dmin, dmax)
        alpha4 = random.uniform(alpha_min, math.pi / 2 - alpha_min)
        d4 = random.uniform(dmin, dmax)
        return cls(ctx, (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4))

    @property
    def genes(self):
        return self._genes

    @genes.setter
    def genes(self, genes):
        """Cut the genes to the limits and reset the fitness."""
        genes = self.check_and_repair_the_geo_size(*genes)
        self._genes = genes
        self._fitness = None
        self._inputs = []
        self._execution_artifact = {}

    @property
    def inputs(self):
        return self._inputs

    @property
    def execution_artifact(self):
        return self._execution_artifact

    @property
    def fitness(self):
        """Calculate the fitness.

        The fitness corresponds of the percentage of the evaluation map covered
        by the geometry.
        """
        if (
            self._fitness is None
        ):  # TODO: Check that the condition is still correct
            self._fitness = 0
            self._execution_artifact["features"] = []
            self._inputs = []
            geom = self.as_geometry()
            query = self._ctx.geometries_tree.query(geom)
            for other in query:
                other_g = self._ctx.geometries[other]
                try:
                    intersection = geom.intersection(other_g)
                except TypeError as e:
                    raise TypeError(f"{e} : {geom} : {other_g}")
                value = float(self._ctx.values[other])
                if not intersection.is_empty:
                    feature = self._ctx.features[other]
                    area = intersection.area
                    fitness = area * pow(value, 3) + 0.00001
                    self._execution_artifact["features"].append(
                        {
                            "feature": feature.id,
                            "area": area,
                            "fitness": fitness,
                        }
                    )
                    self._inputs.append(feature)
                    self._fitness += fitness
        return self._fitness

    def check_and_repair_the_geo_size(
        self, x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4
    ):
        geo_size = self._ctx.geo_size
        step = 0.01 * math.sqrt(geo_size)
        alpha_min = 0.1
        alpha_max = math.pi / 2 - alpha_min
        xmin, ymin, xmax, ymax = self._ctx.bounds

        if x < xmin:
            x = xmin
        elif x > xmax:
            x = xmax

        if y < ymin:
            y = ymin
        elif y > ymax:
            y = ymax

        if alpha1 < alpha_min:
            alpha1 = alpha_min
        elif alpha1 > alpha_max:
            alpha1 = alpha_max

        if alpha2 < alpha_min:
            alpha2 = alpha_min
        elif alpha2 > alpha_max:
            alpha2 = alpha_max

        if alpha3 < alpha_min:
            alpha3 = alpha_min
        elif alpha3 > alpha_max:
            alpha3 = alpha_max

        if alpha4 < alpha_min:
            alpha4 = alpha_min
        elif alpha4 > alpha_max:
            alpha4 = alpha_max

        # calculate the area
        def calculate_area(
            x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4
        ):
            D1 = math.sqrt(
                pow(d1, 2)
                + pow(d4, 2)
                - 2 * d1 * d4 * math.cos(math.pi / 2 + alpha1 - alpha4)
            )
            D2 = math.sqrt(
                pow(d1, 2)
                + pow(d2, 2)
                - 2 * d1 * d2 * math.cos(math.pi / 2 + alpha2 - alpha1)
            )
            D3 = math.sqrt(
                pow(d2, 2)
                + pow(d3, 2)
                - 2 * d2 * d3 * math.cos(math.pi / 2 + alpha3 - alpha2)
            )
            D4 = math.sqrt(
                pow(d3, 2)
                + pow(d4, 2)
                - 2 * d3 * d4 * math.cos(math.pi / 2 + alpha4 - alpha3)
            )
            P1 = (D1 + d1 + d4) / 2
            S1 = math.sqrt(P1 * (P1 - D1) * (P1 - d1) * (P1 - d4))
            P2 = (D2 + d1 + d2) / 2
            S2 = math.sqrt(P2 * (P2 - D2) * (P2 - d1) * (P2 - d2))
            P3 = (D3 + d2 + d3) / 2
            S3 = math.sqrt(P3 * (P3 - D3) * (P3 - d2) * (P3 - d3))
            P4 = (D4 + d3 + d4) / 2
            S4 = math.sqrt(P4 * (P4 - D4) * (P4 - d3) * (P4 - d4))
            S = S1 + S2 + S3 + S4
            return S

        S = calculate_area(
            x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4
        )
        d1 = d1 * math.sqrt(geo_size / S)
        d2 = d2 * math.sqrt(geo_size / S)
        d3 = d3 * math.sqrt(geo_size / S)
        d4 = d4 * math.sqrt(geo_size / S)

        if d1 < step:
            d1 = step
        if d2 < step:
            d2 = step
        if d3 < step:
            d3 = step
        if d4 < step:
            d4 = step

        return x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4

    def as_geometry(self, proj=None):
        """Return the geometry of the individual."""
        x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4 = self.genes
        geom = Polygon(
            [
                (x + d1 * math.cos(alpha1), y + d1 * math.sin(alpha1)),
                (x - d2 * math.sin(alpha2), y + d2 * math.cos(alpha2)),
                (x - d3 * math.cos(alpha3), y - d3 * math.sin(alpha3)),
                (x + d4 * math.sin(alpha4), y - d4 * math.cos(alpha4)),
            ]
        )
        if proj is not None and self._ctx.proj is not None:
            proj_in = pyproj.Proj(self._ctx.proj)
            proj_out = pyproj.Proj(proj)
            if not proj_in.is_exact_same(proj_out):
                project = partial(
                    pyproj.transform, proj_in, proj_out, always_xy=True
                )
                geom = transform(project, geom)
        return geom

    def crossover_1(self, other_parent):
        (
            x1,
            y1,
            alpha11,
            d11,
            alpha21,
            d21,
            alpha31,
            d31,
            alpha41,
            d41,
        ) = self.genes
        (
            x2,
            y2,
            alpha12,
            d12,
            alpha22,
            d22,
            alpha32,
            d32,
            alpha42,
            d42,
        ) = other_parent.genes
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x2, y2, alpha11, d11, alpha21, d21, alpha31, d31, alpha41, d41
        )
        offspring1 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x1, y1, alpha12, d12, alpha22, d22, alpha32, d32, alpha42, d42
        )
        offspring2 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        return [
            Individual(self._ctx, genes=offspring1),
            Individual(self._ctx, genes=offspring2),
        ]

    def crossover_2(self, other_parent):
        (
            x1,
            y1,
            alpha11,
            d11,
            alpha21,
            d21,
            alpha31,
            d31,
            alpha41,
            d41,
        ) = self.genes
        (
            x2,
            y2,
            alpha12,
            d12,
            alpha22,
            d22,
            alpha32,
            d32,
            alpha42,
            d42,
        ) = other_parent.genes
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x1, y1, alpha11, d12, alpha21, d22, alpha31, d32, alpha41, d42
        )
        offspring1 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x2, y2, alpha12, d11, alpha22, d21, alpha32, d31, alpha42, d41
        )
        offspring2 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        return [
            Individual(self._ctx, genes=offspring1),
            Individual(self._ctx, genes=offspring2),
        ]

    def crossover_3(self, other_parent):
        (
            x1,
            y1,
            alpha11,
            d11,
            alpha21,
            d21,
            alpha31,
            d31,
            alpha41,
            d41,
        ) = self.genes
        (
            x2,
            y2,
            alpha12,
            d12,
            alpha22,
            d22,
            alpha32,
            d32,
            alpha42,
            d42,
        ) = other_parent.genes
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x1, y1, alpha12, d11, alpha22, d21, alpha32, d31, alpha42, d41
        )
        offspring1 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x2, y2, alpha11, d12, alpha21, d22, alpha31, d32, alpha41, d42
        )
        offspring2 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        return [
            Individual(self._ctx, genes=offspring1),
            Individual(self._ctx, genes=offspring2),
        ]

    def mutation_1(self, other_parent):
        """Mutate the position (x, y)."""
        (
            x1,
            y1,
            alpha11,
            d11,
            alpha21,
            d21,
            alpha31,
            d31,
            alpha41,
            d41,
        ) = self.genes
        (
            x2,
            y2,
            alpha12,
            d12,
            alpha22,
            d22,
            alpha32,
            d32,
            alpha42,
            d42,
        ) = other_parent.genes
        # select one value at random
        r = random.randint(0, 1)
        if r == 0:
            x1, x2 = x2, x1
        else:
            y1, y2 = y2, y1
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x1, y1, alpha11, d11, alpha21, d21, alpha31, d31, alpha41, d41
        )
        offspring1 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x2, y2, alpha12, d12, alpha22, d22, alpha32, d32, alpha42, d42
        )
        offspring2 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        return [
            Individual(self._ctx, genes=offspring1),
            Individual(self._ctx, genes=offspring2),
        ]

    def mutation_2(self, other_parent):
        """Mutate the vertices (alpha, d)."""
        (
            x1,
            y1,
            alpha11,
            d11,
            alpha21,
            d21,
            alpha31,
            d31,
            alpha41,
            d41,
        ) = self.genes
        (
            x2,
            y2,
            alpha12,
            d12,
            alpha22,
            d22,
            alpha32,
            d32,
            alpha42,
            d42,
        ) = other_parent.genes
        # select one value at random
        r = random.randint(0, 7)
        if r == 0:
            alpha11, alpha12 = alpha12, alpha11
        elif r == 1:
            d11, d12 = d12, d11
        if r == 2:
            alpha21, alpha22 = alpha22, alpha21
        elif r == 3:
            d21, d22 = d22, d21
        if r == 4:
            alpha31, alpha32 = alpha32, alpha31
        elif r == 5:
            d31, d32 = d32, d31
        if r == 6:
            alpha41, alpha42 = alpha42, alpha41
        else:
            d41, d42 = d42, d41
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x1, y1, alpha11, d11, alpha21, d21, alpha31, d31, alpha41, d41
        )
        offspring1 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x2, y2, alpha12, d12, alpha22, d22, alpha32, d32, alpha42, d42
        )
        offspring2 = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)
        return [
            Individual(self._ctx, genes=offspring1),
            Individual(self._ctx, genes=offspring2),
        ]

    def local_search(self):
        # TODO: It doesn't seem to always improve the solution, looks more like
        # a mutation
        x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4 = self.genes
        xmin, ymin, xmax, ymax = self._ctx.bounds
        r = random.randint(0, 10)
        if r == 0:
            x += random.uniform(-(xmax - xmin) / 100, (xmax - xmin) / 100)
        elif r == 1:
            y += random.uniform(-(ymax - ymin) / 100, (ymax - ymin) / 100)
        elif r == 2:
            alpha1 += random.uniform(-math.pi / 8, math.pi / 8)
        elif r == 3:
            d1 += random.uniform(
                -(xmax - xmin + ymax - ymin) / 500,
                (xmax - xmin + ymax - ymin) / 500,
            )
        elif r == 4:
            alpha2 += random.uniform(-math.pi / 8, math.pi / 8)
        elif r == 5:
            d2 += random.uniform(
                -(xmax - xmin + ymax - ymin) / 500,
                (xmax - xmin + ymax - ymin) / 500,
            )
        elif r == 6:
            alpha3 += random.uniform(-math.pi / 8, math.pi / 8)
        elif r == 7:
            d3 += random.uniform(
                -(xmax - xmin + ymax - ymin) / 500,
                (xmax - xmin + ymax - ymin) / 500,
            )
        elif r == 8:
            alpha4 += random.uniform(-math.pi / 8, math.pi / 8)
        elif r == 9:
            d4 += random.uniform(
                -(xmax - xmin + ymax - ymin) / 500,
                (xmax - xmin + ymax - ymin) / 500,
            )
        else:
            rotation = random.uniform(-math.pi / 8, math.pi / 8)
            alpha1 += rotation
            alpha2 += rotation
            alpha3 += rotation
            alpha4 += rotation

        (
            x,
            y,
            alpha1,
            d1,
            alpha2,
            d2,
            alpha3,
            d3,
            alpha4,
            d4,
        ) = self.check_and_repair_the_geo_size(
            x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4
        )
        self.genes = (x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4)


class Population:
    def __init__(self, ctx, individuals):
        self._ctx = ctx
        self.individuals = individuals

    @classmethod
    def random(cls, ctx, size):
        """Initialize a population with random individuals.

        Args:
            size: the size of the population to generate
        """
        individuals = [Individual.random(ctx) for i in range(size)]
        return cls(ctx, individuals)

    def select_roulette(self, count, replace=False):
        """Select individuals in the population using the roulette wheel
        policy.

        The probability of selecting an individual depends on its fitness
        value.
        An invidiuals with a high fitness will have more chance to be selected
        than an individual with a low fitness.

        Args:
            count: the number of individuals to select
            replace: allow to select several times the same individual
                (default: False)

        Returns:
            A new population with selected individuals
        """
        fitness = [
            i.fitness if i.fitness != 0 else 0.00001 for i in self.individuals
        ]
        p = [f / sum(fitness) for f in fitness]
        selection = list(
            np.random.choice(
                self.individuals, size=count, replace=replace, p=p
            )
        )
        copies = [Individual(self._ctx, ind.genes) for ind in selection]
        return Population(self._ctx, copies)

    def combine(self, individuals):
        """Add offsptring to parent population."""
        for offspring in individuals:
            self.individuals.append(offspring)

    def create_children_from_crossover_1(self, count):
        children = []
        parents_1_pop = self.select_roulette(int(count / 2))
        for p1 in parents_1_pop.individuals:
            p2 = np.random.choice(self.individuals)
            children.extend(p1.crossover_1(p2))
        return Population(self._ctx, children)

    def create_children_from_crossover_2(self, count):
        children = []
        parents_1_pop = self.select_roulette(int(count / 2))
        for p1 in parents_1_pop.individuals:
            p2 = np.random.choice(self.individuals)
            children.extend(p1.crossover_2(p2))
        return Population(self._ctx, children)

    def create_children_from_crossover_3(self, count):
        children = []
        parents_1_pop = self.select_roulette(int(count / 2))
        for p1 in parents_1_pop.individuals:
            p2 = np.random.choice(self.individuals)
            children.extend(p1.crossover_3(p2))
        return Population(self._ctx, children)

    def create_children_from_mutation_1(self, count):
        children = []
        parents_1_pop = self.select_roulette(int(count / 2))
        for p1 in parents_1_pop.individuals:
            p2 = np.random.choice(self.individuals)
            children.extend(p1.mutation_1(p2))
        return Population(self._ctx, children)

    def create_children_from_mutation_2(self, count):
        children = []
        parents_1_pop = self.select_roulette(int(count / 2))
        for p1 in parents_1_pop.individuals:
            p2 = np.random.choice(self.individuals)
            children.extend(p1.mutation_2(p2))
        return Population(self._ctx, children)

    def create_children_from_local_search(self, count):
        """One parent and one offspring."""
        offspring = []
        parents = self.select_roulette(count)
        for p in parents.individuals:
            p.local_search()
            offspring.append(p)
        return Population(self._ctx, offspring)


class GeneticAlgorithm:
    """Genetic algorithm to find the best areas on the decision map.

    Attributes:
        population: The population of solution of the genetic algorithm.
        size (int): The size of the population.
        iteration (int): The current iteration of the algorithm.
    """

    def __init__(
        self, ctx, size, mutation=None, children=None, filter_clusters=True
    ):
        self._ctx = ctx
        self.size = size
        self.population = Population.random(ctx, size)
        self.iteration = 0
        self.crossover_1_rate = int(0.3 * size)
        self.crossover_2_rate = int(0.3 * size)
        self.crossover_3_rate = int(0.5 * size)
        self.mutation_1_rate = int(0.3 * size)
        self.mutation_2_rate = int(0.3 * size)
        self.local_search_rate = int(0.5 * size)
        self.filter_clusters = filter_clusters
        self.B = []
        self.all_solutions_found = []

    def filter_best(self):
        def search(Fscore_list, num=3):
            if num > len(set(Fscore_list)):
                num = len(set(Fscore_list))
            res = dict(
                [(v, []) for v in sorted(set(Fscore_list), reverse=True)[:num]]
            )
            for index, val in enumerate(Fscore_list):
                if val in res:
                    res[val].append(index)
            A = sorted(res.items(), key=lambda x: x[0], reverse=True)
            B = A[num - 1]
            return B[0]

        threshold = search(self._ctx.df.value, num=3)  # TODO: Modify that?
        A = []
        for H in self.all_solutions_found:
            check_fitness = H.fitness
            if check_fitness >= 0.9 * self._ctx.geo_size * pow(
                int(threshold), 3
            ):  # threshold of objective function
                A.append(H)
        R = []
        le = len(A)
        max_num = 1
        while le >= 1:
            S = sorted(A, key=lambda i: i.fitness, reverse=True)
            S0 = S[0]  # best solution
            (
                x0,
                y0,
                alpha01,
                d01,
                alpha02,
                d02,
                alpha03,
                d03,
                alpha04,
                d04,
            ) = S0.genes
            B = []
            for M in S:
                x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4 = M.genes
                if math.sqrt(
                    pow(x - x0, 2) + pow(y - y0, 2)
                ) > 0.8 * math.sqrt(
                    self._ctx.geo_size
                ):  # set the distance constraint
                    B.append(M)
            A = B
            R.append(S0)  # save the best solution
            le = len(B)
            max_num += 1
            if max_num >= 30:  # max number of solutions to be shown
                break
        return Population(self._ctx, R)

    def iterate(self):
        """Iterate the algorithm to the next generation."""
        children_from_crossover_1 = (
            self.population.create_children_from_crossover_1(
                self.crossover_1_rate
            )
        )
        children_from_crossover_2 = (
            self.population.create_children_from_crossover_2(
                self.crossover_2_rate
            )
        )
        children_from_crossover_3 = (
            self.population.create_children_from_crossover_3(
                self.crossover_3_rate
            )
        )
        children_from_mutation_1 = (
            self.population.create_children_from_mutation_1(
                self.mutation_1_rate
            )
        )
        children_from_mutation_2 = (
            self.population.create_children_from_mutation_2(
                self.mutation_2_rate
            )
        )
        children_from_local_search = (
            self.population.create_children_from_local_search(
                self.local_search_rate
            )
        )
        self.population.combine(
            children_from_crossover_1.individuals
        )  # add chirldren from crossover 1
        self.population.combine(children_from_crossover_2.individuals)
        self.population.combine(children_from_crossover_3.individuals)
        self.population.combine(children_from_mutation_1.individuals)
        self.population.combine(children_from_mutation_2.individuals)
        self.population.combine(children_from_local_search.individuals)

        best = max(self.population.individuals, key=lambda i: i.fitness)

        # adaptative restart process
        A = best.fitness
        self.B.append(A)
        B_length = len(self.B)
        if B_length > 3:
            if (
                self.B[B_length - 1] == self.B[B_length - 2]
                and self.B[B_length - 1] == self.B[B_length - 3]
                and self.B[B_length - 1] == self.B[B_length - 4]
            ):
                # reinitialize if the best doesn't improve
                self.population = Population.random(self._ctx, self.size - 1)
            else:
                self.population = self.population.select_roulette(
                    self.size - 1
                )
        else:
            self.population = self.population.select_roulette(self.size - 1)
        self.population.individuals.append(best)  # keep the best

        (
            x0,
            y0,
            alpha01,
            d01,
            alpha02,
            d02,
            alpha03,
            d03,
            alpha04,
            d04,
        ) = best.genes
        for i, K in enumerate(self.population.individuals):
            x, y, alpha1, d1, alpha2, d2, alpha3, d3, alpha4, d4 = K.genes
            if math.sqrt(pow(x - x0, 2) + pow(y - y0, 2)) > math.sqrt(
                self._ctx.geo_size
            ):
                self.all_solutions_found.append(K)  # keep all solutions

        self.all_solutions_found.append(best)
        self.iteration += 1



"""Split geometry models from preference models

Revision ID: d1a3e922d5d0
Revises: 9a7260e6a3bf
Create Date: 2025-02-12 15:57:55.461329

"""

import sqlalchemy as sa
from alembic import op

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "d1a3e922d5d0"
down_revision = "9a7260e6a3bf"
branch_labels = None
depends_on = None


old_preftype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    name="preftype",
)
new_preftype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    name="preftype",
)
tmp_preftype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    name="_preftype",
)


old_keepoverlap = sa.Enum(
    "all",
    "min",
    "max",
    "sum",
    "average",
    name="keepoverlap",
)
new_keepoverlap = sa.Enum(
    "min",
    "max",
    "sum",
    "average",
    name="keepoverlap",
)
tmp_keepoverlap = sa.Enum(
    "all",
    "min",
    "max",
    "sum",
    "average",
    name="_keepoverlap",
)


project_root = sa.Table(
    "project_root",
    sa.MetaData(),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("project_root_data_id", sa.Integer(), nullable=False),
)


project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)


input_assoc = sa.Table(
    "data_input_association",
    sa.MetaData(),
    sa.Column("project_data_id", sa.Integer(), nullable=False),
    sa.Column("input_data_id", sa.Integer(), nullable=False),
)


preference_model = sa.Table(
    "preference_model",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("pref_type", new_preftype, nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("data_generator_id", sa.Integer(), nullable=False),
    sa.Column("cut_to_extent", sa.Boolean(), nullable=True),
    sa.Column(
        "keep_overlap",
        sa.Enum("all", "min", "max", "sum", "average", name="keepoverlap"),
        nullable=False,
    ),
    sa.Column("radius", sa.BigInteger(), nullable=True),
    sa.Column("majority_threshold", sa.Float(), nullable=True),
    sa.Column("rule", sa.String(), nullable=True),
)

old_preference_model = sa.Table(
    "preference_model",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("pref_type", new_preftype, nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("data_generator_id", sa.Integer(), nullable=False),
    sa.Column(
        "keep_overlap",
        sa.Enum("all", "min", "max", "sum", "average", name="keepoverlap"),
        nullable=False,
    ),
    sa.Column("cut_to_extent", sa.Boolean(), nullable=True),
    sa.Column("dissolve_adjacent", sa.Boolean(), nullable=True),
    sa.Column("radius", sa.BigInteger(), nullable=True),
    sa.Column("majority_threshold", sa.Float(), nullable=True),
    sa.Column("rule", sa.String(), nullable=True),
)


def upgrade():
    # Replace old preftype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_preftype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN pref_type TYPE _preftype"
        " USING pref_type::text::_preftype"
    )
    old_preftype.drop(op.get_bind(), checkfirst=False)
    new_preftype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN pref_type TYPE preftype"
        " USING pref_type::text::preftype"
    )
    tmp_preftype.drop(op.get_bind(), checkfirst=False)

    conn = op.get_bind()

    nb_migration_ops = conn.execute(
        sa.text(
            "SELECT count(pdata.id) FROM project_data AS pdata "
            "JOIN preference_model as pref ON pdata.active_model_id = pref.id "
            "WHERE pref.dissolve_adjacent = true"
        )
    ).fetchone()[0]
    nb_migration_ops += conn.execute(
        sa.text(
            "SELECT count(pdata.id) FROM project_data AS pdata "
            "JOIN preference_model as pref ON pdata.active_model_id = pref.id "
            "WHERE pref.keep_overlap != 'all'"
        )
    ).fetchone()[0]
    op_it = 0
    progress_bar(op_it, nb_migration_ops)

    # Get all project data with active model with dissolve adjacent to true
    res = conn.execute(
        sa.select(
            project_data.c[
                "id", "project_id", "name", "data_id", "last_update"
            ],
            old_preference_model.c["cut_to_extent"],
            project_root.c["project_root_data_id"],
        )
        .join(
            old_preference_model,
            project_data.c.active_model_id == old_preference_model.c.id,
        )
        .join(
            project_root,
            project_root.c.project_id == project_data.c.project_id,
        )
        .where(
            old_preference_model.c.dissolve_adjacent
            == True  # noqa: E712, E501
        )
    )
    for (
        pdata_id,
        project_id,
        name,
        data_id,
        last_update,
        cut_to_e,
        project_root_id,
    ) in res:
        # Insert new empty data generator in project
        # And give it current generated data
        new_pdata = conn.execute(
            project_data.insert()
            .values(
                data_type="generator",
                project_id=project_id,
                name=name,
                data_id=data_id,
                last_update=last_update,
            )
            .returning(project_data.c.id)
        ).first()
        # Insert new dissolve_adjacent model into created data generator
        new_model = conn.execute(
            preference_model.insert()
            .values(
                pref_type="dissolve_adjacent",
                data_generator_id=new_pdata[0],
                cut_to_extent=cut_to_e,
                keep_overlap="all",
            )
            .returning(preference_model.c.id)
        ).first()
        # Make dissolve adjacent model active
        conn.execute(
            project_data.update()
            .where(project_data.c.id == new_pdata[0])
            .values(active_model_id=new_model[0])
        )
        # Detach generated data from old project data
        # Reset last_update of current project data which data is moved
        conn.execute(
            project_data.update()
            .where(project_data.c.id == pdata_id)
            .values(data_id=None, last_update=None)
        )
        # Rewire all project data inputs from project data to new data
        # generator
        conn.execute(
            input_assoc.update()
            .where(input_assoc.c.input_data_id == pdata_id)
            .values(input_data_id=new_pdata[0])
        )
        # Add project data as an input to new data generator
        conn.execute(
            input_assoc.insert().values(
                project_data_id=new_pdata[0], input_data_id=pdata_id
            )
        )
        if pdata_id == project_root_id:
            # Data is root of project => update to new project data
            conn.execute(
                project_root.update()
                .where(project_root.c.project_id == project_id)
                .values(project_root_data_id=new_pdata[0])
            )
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    # Remove dissolve_adjacent preference model column
    with op.batch_alter_table("preference_model", schema=None) as batch_op:
        batch_op.drop_column("dissolve_adjacent")

    res = conn.execute(
        sa.select(
            project_data.c[
                "id", "project_id", "name", "data_id", "last_update"
            ],
            old_preference_model.c["cut_to_extent", "keep_overlap"],
            project_root.c["project_root_data_id"],
        )
        .join(
            old_preference_model,
            project_data.c.active_model_id == old_preference_model.c.id,
        )
        .join(
            project_root,
            project_root.c.project_id == project_data.c.project_id,
        )
        .where(old_preference_model.c.keep_overlap != "all")
    )
    for (
        pdata_id,
        project_id,
        name,
        data_id,
        last_update,
        cut_to_e,
        keep_overlap,
        project_root_id,
    ) in res:
        # Insert new empty data generator in project
        # And give it current generated data
        new_pdata = conn.execute(
            project_data.insert()
            .values(
                data_type="generator",
                project_id=project_id,
                name=name,
                data_id=data_id,
                last_update=last_update,
            )
            .returning(project_data.c.id)
        ).first()
        # Insert new merge_overlap model into created data generator
        new_model = conn.execute(
            preference_model.insert()
            .values(
                pref_type="merge_overlap",
                data_generator_id=new_pdata[0],
                cut_to_extent=cut_to_e,
                keep_overlap=keep_overlap,
            )
            .returning(preference_model.c.id)
        ).first()
        # Make merge overlap model active
        conn.execute(
            project_data.update()
            .where(project_data.c.id == new_pdata[0])
            .values(active_model_id=new_model[0])
        )
        # Detach generated data from old project data
        # Reset last_update of current project data which data is moved
        conn.execute(
            project_data.update()
            .where(project_data.c.id == pdata_id)
            .values(data_id=None, last_update=None)
        )
        # Rewire all project data inputs from project data to new data
        # generator
        conn.execute(
            input_assoc.update()
            .where(input_assoc.c.input_data_id == pdata_id)
            .values(input_data_id=new_pdata[0])
        )
        # Add project data as an input to new data generator
        conn.execute(
            input_assoc.insert().values(
                project_data_id=new_pdata[0], input_data_id=pdata_id
            )
        )
        if pdata_id == project_root_id:
            # Data is root of project => update to new project data
            conn.execute(
                project_root.update()
                .where(project_root.c.project_id == project_id)
                .values(project_root_data_id=new_pdata[0])
            )
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    # Remove all keep_overlap 'all' values, replace with default
    op.execute(
        preference_model.update()
        .where(preference_model.c.keep_overlap == "all")
        .values(keep_overlap="max")
    )

    # Replace old keepoverlap with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_keepoverlap.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN keep_overlap "
        "TYPE _keepoverlap USING keep_overlap::text::_keepoverlap"
    )
    old_keepoverlap.drop(op.get_bind(), checkfirst=False)
    new_keepoverlap.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN keep_overlap "
        "TYPE keepoverlap USING keep_overlap::text::keepoverlap"
    )
    tmp_keepoverlap.drop(op.get_bind(), checkfirst=False)


def downgrade():
    # Replace new keepoverlap with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_keepoverlap.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN keep_overlap "
        "TYPE _keepoverlap USING keep_overlap::text::_keepoverlap"
    )
    new_keepoverlap.drop(op.get_bind(), checkfirst=False)
    old_keepoverlap.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN keep_overlap "
        "TYPE keepoverlap USING keep_overlap::text::keepoverlap"
    )
    tmp_keepoverlap.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("preference_model", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "dissolve_adjacent",
                sa.BOOLEAN(),
                autoincrement=False,
                nullable=True,
                default=False,
            )
        )

    # Replace keep_overlap with default 'all' in all but merge_overlap models
    op.execute(
        old_preference_model.update()
        .where(
            old_preference_model.c.pref_type != "merge_overlap"
        )  # noqa: E711
        .values(keep_overlap="all")
    )

    conn = op.get_bind()

    nb_migration_ops = conn.execute(
        sa.text(
            "SELECT count(pdata.id) FROM project_data AS pdata "
            "JOIN preference_model as pref ON pdata.active_model_id = pref.id "
            "WHERE pref.pref_type IN ('dissolve_adjacent', 'merge_overlap')"
        )
    ).fetchone()[0]
    op_it = 0
    progress_bar(op_it, nb_migration_ops)

    # Get all project data with active model dissolve adjacent
    res = conn.execute(
        sa.select(
            project_data.c["id", "last_update", "data_id", "project_id"],
            project_root.c["project_root_data_id"],
        )
        .join(
            old_preference_model,
            project_data.c.active_model_id == old_preference_model.c.id,
        )
        .join(
            project_root,
            project_root.c.project_id == project_data.c.project_id,
        )
        .where(old_preference_model.c.pref_type == "dissolve_adjacent")
    )
    for (
        pdata_id,
        last_update,
        data_id,
        project_id,
        project_root_id,
    ) in res:
        # Get unique input of project data
        input_id = conn.execute(
            sa.select(input_assoc.c["input_data_id"]).where(
                input_assoc.c.project_data_id == pdata_id
            )
        ).fetchone()[0]

        # Get every other output of project data's input
        other_outputs_of_input = conn.execute(
            input_assoc.select().where(
                input_assoc.c.input_data_id == input_id,
                input_assoc.c.project_data_id != pdata_id,
            )
        ).all()
        # Rewire all project data outputs to input data
        conn.execute(
            input_assoc.update()
            .where(input_assoc.c.input_data_id == pdata_id)
            .values(input_data_id=input_id)
        )
        # Remove link between project data and its input
        conn.execute(
            input_assoc.delete().where(
                input_assoc.c.project_data_id == pdata_id
            )
        )
        if pdata_id == project_root_id:
            # Data is root of project => update to new project data
            conn.execute(
                project_root.update()
                .where(project_root.c.project_id == project_id)
                .values(project_root_data_id=input_id)
            )

        if len(other_outputs_of_input) == 0:
            # project data is the only output of its input data
            # We can merge them
            input_model = conn.execute(
                sa.select(
                    old_preference_model.c["id"],
                )
                .join(
                    project_data,
                    old_preference_model.c.data_generator_id
                    == project_data.c.id,
                )
                .where(
                    old_preference_model.c.data_generator_id == input_id,
                    old_preference_model.c.id
                    == project_data.c.active_model_id,
                )
            ).first()
            # Make input data active model dissolve adjacent
            conn.execute(
                old_preference_model.update()
                .where(old_preference_model.c.id == input_model[0])
                .values(dissolve_adjacent=True)
            )
            # Move generated data from project data to input data
            conn.execute(
                project_data.update()
                .where(project_data.c.id == input_id)
                .values(data_id=data_id, last_update=last_update)
            )
        # Set project data active model to null
        # So to remove references to project_data
        conn.execute(
            project_data.update()
            .where(project_data.c.id == pdata_id)
            .values(active_model_id=None)
        )
        # Remove project_data models
        conn.execute(
            old_preference_model.delete().where(
                old_preference_model.c.data_generator_id == pdata_id
            )
        )
        # Remove project data
        conn.execute(
            project_data.delete().where(project_data.c.id == pdata_id)
        )
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    # Get all project data with an active merge_overlap model
    res = conn.execute(
        sa.select(
            project_data.c["id", "last_update", "data_id", "project_id"],
            old_preference_model.c["keep_overlap", "dissolve_adjacent"],
            project_root.c["project_root_data_id"],
        )
        .join(
            old_preference_model,
            project_data.c.active_model_id == old_preference_model.c.id,
        )
        .join(
            project_root,
            project_root.c.project_id == project_data.c.project_id,
        )
        .where(old_preference_model.c.pref_type == "merge_overlap")
    )
    for (
        pdata_id,
        last_update,
        data_id,
        project_id,
        keep_overlap,
        dissolve_adjacent,
        project_root_id,
    ) in res:
        # Get unique input of project data
        input_id = conn.execute(
            sa.select(input_assoc.c["input_data_id"]).where(
                input_assoc.c.project_data_id == pdata_id
            )
        ).fetchone()[0]

        # Get every other output of project data's input
        other_outputs_of_input = conn.execute(
            input_assoc.select().where(
                input_assoc.c.input_data_id == input_id,
                input_assoc.c.project_data_id != pdata_id,
            )
        ).all()
        # Rewire all project data outputs to input data
        conn.execute(
            input_assoc.update()
            .where(input_assoc.c.input_data_id == pdata_id)
            .values(input_data_id=input_id)
        )
        # Remove link between project data and its input
        conn.execute(
            input_assoc.delete().where(
                input_assoc.c.project_data_id == pdata_id
            )
        )
        if pdata_id == project_root_id:
            # Data is root of project => update to new project data
            conn.execute(
                project_root.update()
                .where(project_root.c.project_id == project_id)
                .values(project_root_data_id=input_id)
            )

        if len(other_outputs_of_input) == 0:
            # project data is the only output of its input data
            # We can merge them
            input_model = conn.execute(
                sa.select(
                    old_preference_model.c["id"],
                )
                .join(
                    project_data,
                    old_preference_model.c.data_generator_id
                    == project_data.c.id,
                )
                .where(
                    old_preference_model.c.data_generator_id == input_id,
                    old_preference_model.c.id
                    == project_data.c.active_model_id,
                )
            ).first()
            # Make input data active model dissolve adjacent
            conn.execute(
                old_preference_model.update()
                .where(old_preference_model.c.id == input_model[0])
                .values(
                    keep_overlap=keep_overlap,
                    dissolve_adjacent=dissolve_adjacent,
                )
            )
            # Move generated data from project data to input data
            conn.execute(
                project_data.update()
                .where(project_data.c.id == input_id)
                .values(data_id=data_id, last_update=last_update)
            )
        # Set project data active model to null
        # So to remove references to project_data
        conn.execute(
            project_data.update()
            .where(project_data.c.id == pdata_id)
            .values(active_model_id=None)
        )
        # Remove project_data models
        conn.execute(
            old_preference_model.delete().where(
                old_preference_model.c.data_generator_id == pdata_id
            )
        )
        # Remove project data
        conn.execute(
            project_data.delete().where(project_data.c.id == pdata_id)
        )
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    # Replace new preftype with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_preftype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN pref_type "
        "TYPE _preftype USING pref_type::text::_preftype"
    )
    new_preftype.drop(op.get_bind(), checkfirst=False)
    old_preftype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN pref_type "
        "TYPE preftype USING pref_type::text::preftype"
    )
    tmp_preftype.drop(op.get_bind(), checkfirst=False)



from .geo_data import *  # noqa: F401,F403
from .mrsort_inference import *  # noqa: F401,F403
from .processing_models import *  # noqa: F401,F403
from .project import *  # noqa: F401,F403
from .project_data import *  # noqa: F401,F403
from .requests import *  # noqa: F401,F403
from .schemas import *  # noqa: F401,F403
from .share import *  # noqa: F401,F403



from datetime import datetime

from passlib.apps import custom_app_context as pwd_context

from ..exceptions import UserTokenError
from ..security import jwt_handler
from . import db
from .mixins import ModelMixin


class UserRefreshToken(ModelMixin, db.Model):
    __tablename__ = "user_refresh_token"

    user_id = db.Column(db.ForeignKey("user.id"))
    refresh_token = db.Column(db.LargeBinary)
    expiration = db.Column(db.DateTime)

    user = db.relationship("User", back_populates="tokens")

    def __init__(self, token):
        self.refresh_token = token
        self.expiration = datetime.fromtimestamp(
            jwt_handler.get_token_exp(token)
        )


class User(ModelMixin, db.Model):
    """Represents a user"""

    __tablename__ = "user"

    username = db.Column(db.String, unique=True)
    email = db.Column(db.String, unique=True)
    password_hash = db.Column(db.String)

    tokens = db.relationship(
        "UserRefreshToken",
        single_parent=True,
        back_populates="user",
        cascade="all,delete-orphan",
    )

    def __init__(self, username, email, password=None):
        """Initialise the user"""
        self.username = username
        self.email = email
        if password is not None:
            self.set_password(password)
        self.tokens = []

    def __repr__(self):
        return "<User(username={}, email={}>".format(self.username, self.email)

    @classmethod
    def get_by_username(cls, username):
        return cls.query.filter_by(username=username).first()

    @classmethod
    def get_by_email(cls, email):
        return cls.query.filter_by(email=email).first()

    @classmethod
    def get_from_jwt(cls, token):
        user_id = jwt_handler.get_user_id(token)
        if user_id is not None:
            return cls.get_by_id(user_id)
        else:
            raise UserTokenError(
                "No valid user found for the token {}".format(token)
            )

    @classmethod
    def get_from_refresh_token(cls, token):
        user_id = jwt_handler.get_user_id(token)
        if user_id is not None:
            user = cls.get_by_id(user_id)
            if user.check_refresh_token(token) is False:
                raise UserTokenError(
                    "The refresh token can be read but is not valid"
                )
            else:
                return user
        else:
            raise UserTokenError("No found for the token {}".format(token))

    def set_password(self, password):
        """Set the password_hash from the given password"""
        self.password_hash = pwd_context.encrypt(password)

    def check_password(self, password):
        """Check if the password is correct"""
        return pwd_context.verify(password, self.password_hash)

    def create_jwt(self):
        """Create a JSON Web Token for this user"""
        token = jwt_handler.create_user_token(self.id)
        return token

    def create_refresh_token(self):
        """Creates a JSON refresh token, and associates it to the user"""
        refresh = jwt_handler.create_user_refresh_token(self.id)
        self.tokens.append(UserRefreshToken(refresh.encode()))
        return refresh

    def check_refresh_token(self, refresh_token):
        """Checks if the refresh is saved in the database for the current user.

        WARNING: Does not check the validity of the token, checks only if it
        exists for the current user

        Args:
            refresh_token - the token to check. Can be a str or a bytes object
        """
        for token in self.tokens:
            try:
                if token.refresh_token == refresh_token.encode():
                    return True
            except AttributeError:
                if token.refresh_token == refresh_token:
                    return True
        return False



from flask import Flask


def init_app(app: Flask):
    from .db import register_commands as register_db_commands
    from .user import user_cli

    app.cli.add_command(user_cli)
    register_db_commands(app)



import logging
import os

from flask import Flask, jsonify
from marshmallow import ValidationError
from werkzeug.exceptions import default_exceptions
from werkzeug.middleware.profiler import ProfilerMiddleware

from .exceptions import RequestError
from .models import db
from .resources import api
from .security import jwt_handler


def create_app(instance_config_file=None) -> Flask:
    instance_path = os.path.abspath(os.getcwd())
    app = Flask(
        __name__,
        instance_relative_config=True,
        instance_path=instance_path,
        static_folder=None,
    )

    handler = logging.StreamHandler()
    handler.setLevel(logging.WARNING)
    app.logger.addHandler(handler)

    app.config.from_object("deseasion.backend.config")
    if instance_config_file:
        app.config.from_pyfile(instance_config_file, silent=True)

    if app.config.get("PROFILE", False):
        app.wsgi_app = ProfilerMiddleware(
            app.wsgi_app, profile_dir=app.config["PROFILE_DIR"]
        )

    db.init_app(app)
    api.init_app(app)
    jwt_handler.init_app(app)

    @app.errorhandler(RequestError)
    def handler_request_error(error):
        app.logger.info("Request error: {}".format(str(error)), exc_info=True)
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response

    @app.errorhandler(ValidationError)
    def handler_validation_error(error):
        app.logger.info(
            "Validation error: {}".format(str(error)), exc_info=True
        )
        response = jsonify(
            {"message": "Validation error", "errors": error.messages}
        )
        response.status_code = 400
        return response

    def handler_http_exception(error):
        response = jsonify({"message": error.description, "error": str(error)})
        response.status_code = error.code
        return response

    # trap the default http exceptions (HTTPException subclasses)
    # https://github.com/pallets/flask/issues/941
    for code in default_exceptions:
        app.register_error_handler(code, handler_http_exception)

    @app.errorhandler(Exception)
    def handler_exception(error):
        app.logger.warning("Unhandled exception:", exc_info=True)
        response = jsonify({"message": "Internal Server Error"})
        response.status_code = 500
        return response

    return app



"""add alternative class parameters to WMS geodata

Revision ID: 4ca4e5e44ff2
Revises: 2ce65db29472
Create Date: 2024-12-09 17:18:21.148120

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4ca4e5e44ff2"
down_revision = "2ce65db29472"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "DELETE FROM geo_feature WHERE id in "
        "(SELECT id FROM feature WHERE data_id IS NULL)"
    )
    op.execute("DELETE FROM feature WHERE data_id IS NULL")
    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.alter_column(
            "data_id", existing_type=sa.INTEGER(), nullable=False
        )

    with op.batch_alter_table("wms_geo_data", schema=None) as batch_op:
        batch_op.add_column(sa.Column("start", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("step", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("stop", sa.Float(), nullable=True))


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("wms_geo_data", schema=None) as batch_op:
        batch_op.drop_column("stop")
        batch_op.drop_column("step")
        batch_op.drop_column("start")

    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.alter_column(
            "data_id", existing_type=sa.INTEGER(), nullable=True
        )

    # ### end Alembic commands ###



"""Add creation/modification timestamps to ProjectData

Revision ID: 141e705ab129
Revises: 5f84e56d0465f
Create Date: 2025-04-01 14:24:30.773709

"""

from datetime import datetime, timezone

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "141e705ab129"
down_revision = "5f84e56d0465f"
branch_labels = None
depends_on = None


project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)


def upgrade():
    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("created_at", sa.DateTime(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("modified_at", sa.DateTime(), nullable=True)
        )

    # Create dummy creation date to last update
    op.execute(sa.text("UPDATE project_data SET created_at = last_update"))

    # Create dummy modified date to last update
    op.execute(sa.text("UPDATE project_data SET modified_at = last_update"))

    # Change modified date of generators so they are outdated
    now = datetime.now(timezone.utc)
    op.execute(
        project_data.update()
        .values(modified_at=now)
        .where(project_data.c.data_type == "generator")
    )


def downgrade():
    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_column("modified_at")
        batch_op.drop_column("created_at")



from sqlalchemy.orm import joinedload
from sqlalchemy.sql import text

from ..exceptions import PermissionError, RequestError
from ..models import (
    BaseData,
    BaseGeoData,
    GeneratedGeoData,
    GeoData,
    GeoDataType,
    GlobalData,
    StreamGeoData,
    db,
)
from ..models.geo_data import UploadableData
from .permission_service import has_permission


class DataService:
    __model__ = BaseData

    def has_permission(self, data, user=None):
        """
        Return if the current user is authorized to access the data.

        Args:
            data: A data object.

        Returns: authorization as bool
        """
        if isinstance(data, UploadableData):
            return has_permission(data, user)
        elif isinstance(data, GeneratedGeoData) and data.project_data:
            return has_permission(data.project_data.project, user)
        return False

    def check_permission(self, data):
        """
        Check if the current user is authorized to access the data.

        Args:
            data: A data object.

        Raises:
            PermissionError is the user is not authorized to use the data.
        """
        if not self.has_permission(data):
            raise PermissionError

    def get_if_authorized(self, data_id):
        """
        Return the global data if the user is authorized to access it.

        Args:
            data_id (int): The id of the data to return.

        Raises:
            RequestError (code 404) if the data does not exist.

        Returns:
            The data.
        """
        data = BaseData.get_by_id(data_id)
        if data is None:
            raise RequestError("The data does not exist", 404)
        self.check_permission(data)
        return data

    def get_all_authorized_data(self, user, types: set[str]):
        """
        Return the list of all the data.

        Args:
            user: User for which to get the data.

        Returns:
            A list of `Data` objects.
        """
        data_list = (
            BaseData.query.where(BaseData.type.in_(types))
            .order_by(BaseData.id)
            .all()
        )
        return list(filter(lambda d: self.has_permission(d, user), data_list))


class GlobalDataService:
    __model__ = GlobalData

    def check_permission(self, data):
        """
        Check if the current user is authorized to access the global data.

        Args:
            data: A global data object.

        Raises:
            PermissionError is the user is not authorized to use the global
            data.
        """
        if isinstance(data, UploadableData):
            if not has_permission(data):
                raise PermissionError

    def get_if_authorized(self, data_id):
        """
        Return the global data if the user is authorized to access it.

        Args:
            data_id (int): The id of the global data to return.

        Raises:
            RequestError (code 404) if the global data does not exist.

        Returns:
            The data.
        """
        data = GlobalData.get_by_id(data_id)
        if data is None:
            raise RequestError("The global data does not exist", 404)
        self.check_permission(data)
        return data

    def get_all_authorized_data(self, user):
        """
        Return the list of all the global data.

        Args:
            user: User for which to get the global data.

        Returns:
            A list of `GlobalData` objects.
        """
        data_list = (
            GlobalData.query.options(joinedload(GlobalData.permissions))
            .order_by(GlobalData.id)
            .all()
        )
        return list(filter(lambda g: has_permission(g, user), data_list))


class StreamGeoDataService:
    __model__ = StreamGeoData

    def check_permission(self, data):
        """
        Check if the current user is authorized to access the data.

        Args:
            data: A geo-data stream object.

        Raises:
            PermissionError is the user is not authorized to use the data.
        """
        if isinstance(data, StreamGeoData):
            if not has_permission(data):
                raise PermissionError

    def get_if_authorized(self, stream_geo_data_id):
        """
        Return the geo data stream if the user is authorized to access it.

        Args:
            stream_geo_data_id (int): The id of the geo data stream to return.

        Raises:
            RequestError (code 404) if the stream does not exist.

        Returns:
            The geo data stream.
        """
        data = BaseData.get_by_id(stream_geo_data_id)
        if not isinstance(data, StreamGeoData):
            raise RequestError("The geo data stream does not exist", 404)
        self.check_permission(data)
        return data

    def get_all_authorized_geo_data(self, user):
        """
        Return the list of all the geo data streams.

        Args:
            user: User for which to get the geo data streams.

        Returns:
            A list of `StreamGeoData` objects.
        """
        data_list = (
            BaseData.query.options(joinedload(BaseData.permissions))
            .where(BaseData.type.in_([GeoDataType.wfs, GeoDataType.wms]))
            .order_by(BaseData.id)
            .all()
        )
        return list(filter(lambda g: has_permission(g, user), data_list))


class GeoDataService:
    __model__ = BaseGeoData

    def check_permission(self, data):
        """
        Check if the current user is authorized to access the data.

        Args:
            data: A geo-data object.

        Raises:
            PermissionError is the user is not authorized to use the data.
        """
        if isinstance(data, UploadableData):
            if not has_permission(data):
                raise PermissionError
        elif isinstance(data, GeneratedGeoData):
            if not has_permission(data.project_data.project):
                raise PermissionError

    def get_if_authorized(self, geo_data_id):
        """
        Return the geo data if the user is authorized to access it.

        Args:
            geo_data_id (int): The id of the geo data to return.

        Raises:
            RequestError (code 404) if the data does not exist.

        Returns:
            The geo data.
        """
        data = BaseGeoData.get_by_id(geo_data_id)
        if data is None:
            raise RequestError("The geo data does not exist", 404)
        self.check_permission(data)
        return data

    def get_all_authorized_geo_data(self, user):
        """
        Return the list of all the geo data (not the generated geo data).

        Args:
            user: User for which to get the geo data.

        Returns:
            A list of `GeoData` objects.
        """
        data_list = (
            GeoData.query.options(
                joinedload(GeoData.attributes), joinedload(GeoData.permissions)
            )
            .order_by(GeoData.id)
            .all()
        )
        return list(filter(lambda g: has_permission(g, user), data_list))

    def get_vector_tile(self, geo_data_id, z, x, y, property_name=None):
        """
        Return the geo data features as a Mapbox Vector Tile.

        The geometries are simplified to the zoom level (z coordinate) and
        projected as web mercator. If a `property_name` is specified.
        Geometry near the poles are clipped as they cannot be transformed to
        Web Mercator Projection (EPSG:3857).
        The clipping is done above 85° North and below -85° South.

        Args:
            geo_data_id (int): The id of the geo data to return.
            z (int): Z coordinate of the tile (zoom level).
            x (int): X coordinate of the tile.
            y (int): Y coordinate of the tile.
            property_name (str): The name of a property to return as an
                attribute in the geometries. Default: None.

        Returns:
            A string representing a Mapbox Vector Tile.
        """
        # Calculate the resolution for the zoom level
        # https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Resolution_and_Scale  # noqa: E501
        tolerance = 156543.03 / (2**z)
        # exclude geometry too near the poles:
        # ST_Intersection(geom, ST_MakeEnvelope(-180, -85, 180, 85, 4326))
        # https://gis.stackexchange.com/questions/357847/postgis-vector-tiles-zoom-affects-tile-production  # noqa: E501
        # https://en.wikipedia.org/wiki/Web_Mercator_projection
        if property_name is not None:
            q = (
                "SELECT ST_AsMVT(q) FROM ("
                "SELECT f.id as id, v.value->>0 as value, "
                "ST_AsMVTGeom(ST_SimplifyVW(ST_Transform("
                "ST_Intersection(f.geom, ST_MakeEnvelope("
                "-180, -85, 180, 85, 4326)), 3857), :tolerance), "
                "TileBBox(:z, :x, :y)) AS geom "
                "FROM geo_feature f "
                "INNER JOIN feature bf ON f.id = bf.id "
                "LEFT JOIN data_attribute a ON bf.data_id = a.data_id "
                "LEFT JOIN data_value v ON a.id = v.attribute_id "
                "AND f.id = v.feature_id "
                "WHERE bf.data_id = :geo_data_id AND f.geom && "
                "TileBBox(:z, :x, :y, 4326) "
                "AND a.name = :prop) q"
            )
            keys = {
                "tolerance": tolerance,
                "z": z,
                "x": x,
                "y": y,
                "geo_data_id": geo_data_id,
                "prop": property_name,
            }
        else:
            q = (
                "SELECT ST_AsMVT(q) FROM ("
                "SELECT f.id, ST_AsMVTGeom(ST_SimplifyVW("
                "ST_Transform(ST_Intersection(f.geom, ST_MakeEnvelope("
                "-180, -85, 180, 85, 4326)), 3857), :tolerance), "
                "TileBBox(:z, :x, :y)) AS geom "
                "FROM geo_feature f "
                "INNER JOIN feature bf ON f.id = bf.id "
                "WHERE bf.data_id = :geo_data_id AND f.geom && "
                "TileBBox(:z, :x, :y, 4326)) q"
            )
            keys = {
                "tolerance": tolerance,
                "z": z,
                "x": x,
                "y": y,
                "geo_data_id": geo_data_id,
            }
        result = db.session.execute(text(q), keys)
        mvt = result.first()[0]
        return mvt
